import copy
import httpx
import os

DATA_DIR = os.path.join(os.path.abspath(os.path.dirname(__file__)), 'data')


class GenericSideEffect:

    def __init__(self, data):
        self.orig_data = data
        self.data = copy.deepcopy(self.orig_data)

    def __call__(self, request, route):
        return httpx.Response(200, json=self.data)

    def __enter__(self):
        self.reset()
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.reset()

    def reset(self):
        self.data = copy.deepcopy(self.orig_data)


def attribute_test(response, data, parents=''):
    for k, v in data.items():
        if isinstance(v, dict):
            if parents:
                _parents = f'{parents}.{k}'
            else:
                _parents = k
            attribute_test(response, v, parents=_parents)
        elif isinstance(v, list):
            for idx, child in enumerate(v):
                if isinstance(child, dict):
                    if parents:
                        _parents = f'{parents}.{k}[{idx}]'
                    else:
                        _parents = f'{k}[{idx}]'
                    attribute_test(response, child, parents=_parents)
                else:
                    if parents:
                        assert eval(f'response.{parents}.{k}[{idx}]') == v[idx]
                    else:
                        assert eval(f'response.{k}[{idx}]') == v[idx]
        else:
            if parents:
                assert eval(f'response.{parents}.{k}') == v
            else:
                assert eval(f'response.{k}') == v



import dateutil
import functools
import itertools
import logging

from datetime import datetime, timezone

from snisp import exceptions, utils, systems
from snisp.contracts import Contract
from snisp.decorators import cooldown, docked, in_orbit, retry, transit
from snisp.markets import Markets
from snisp.shipyards import Shipyards
from snisp.systems import System
from snisp.waypoints import Waypoints


logger = logging.getLogger(__name__)


class Fleet:

    """Fleet is the top-level class for accessing Ships"""

    def __init__(self, agent):
        self.agent = agent

    def __repr__(self):
        return f'{self.__class__.__name__}({self.agent!r})'

    @retry()
    def __call__(self, ship_symbol):
        """
        Retrieves a Ship instance

        Args:
            ship_symbol: The symbol for the ship to retrieve

        Returns:
            Ship
        """
        ship_symbol = ship_symbol.upper()
        response = self.agent.client.get(f'/my/ships/{ship_symbol}')
        return Ship(self.agent, response.json()['data'])

    def __iter__(self):
        """
        Iterates over the current Agent's Fleet

        Yields:
            Ship
        """
        page = 1
        response = self.get_page(page=page)
        while data := response.json()['data']:
            for ship in data:
                if ship['symbol'] not in self.agent.dead_ships:
                    yield Ship(self.agent, ship)
            page += 1
            response = self.get_page(page=page)

    @retry()
    def get_page(self, page=1, limit=20):
        params = {'limit': int(limit), 'page': int(page)}
        return self.agent.client.get('/my/ships', params=params)

    def drones(self):
        """Yields Ships where ship.frame.symbol == 'FRAME_DRONE'"""
        for ship in self:
            if ship.frame.symbol == 'FRAME_DRONE':
                yield ship

    def freighters(self):
        """Yields Ships where 'FREIGHTER' in ship.frame.symbol"""
        for ship in self:
            if 'FREIGHTER' in ship.frame.symbol:
                yield ship

    def mining_drones(self):
        """Yields Ships that are Drones and have a Mining Laser Mount"""
        for drone in self.drones():
            for mount in drone.mounts:
                if mount.symbol.upper().startswith('MOUNT_MINING_LASER_'):
                    yield drone

    def siphon_drones(self):
        """Yields Ships that are Drones and have a Gas Siphon MountShip"""
        for drone in self.drones():
            for mount in drone.mounts:
                if mount.symbol.upper().startswith('MOUNT_GAS_SIPHON_'):
                    yield drone

    def probes(self):
        """Yields Ships where ship.frame.symbol == 'FRAME_PROBE'"""
        for ship in self:
            if ship.frame.symbol == 'FRAME_PROBE':
                yield ship

    def ships(self):
        """Yields Ships that are neither Drone nor Probes"""
        for ship in self:
            if 'DRONE' not in ship.frame.symbol:
                if 'PROBE' not in ship.frame.symbol:
                    yield ship

    def shuttles(self):
        """Yields Ships where 'SHUTTLE' in ship.frame.symbol"""
        for ship in self:
            if 'SHUTTLE' in ship.frame.symbol:
                yield ship


class Ship(utils.AbstractJSONItem):

    """A Ship can be a Drone, Probe, Freighter, etc."""

    def __init__(self, agent, ship_data):
        self.agent = agent
        self._data = ship_data

    @property
    def arrival(self):
        """
        Property that returns the number of seconds until the Ship reaches
        the destination

        Blocks:
            False

        Returns:
            int: sceond to arrival if ship is IN_TRANSIT; else, 0
        """
        if self.nav.status == 'IN_TRANSIT':
            arrival = dateutil.parser.parse(self.nav.route.arrival)
            delta = arrival - datetime.now(timezone.utc)
            return delta.seconds if delta.days >= 0 else 0
        return 0

    @property
    def location(self):
        """Property for the Ship's current Location

        Blocks:
            False

        Returns:
            snisp.system.Location: Ship's current Location
        """
        return systems.Location(self.agent, self.to_dict())

    @property
    def waypoints(self):
        """
        Property that returns the snisp.waypoints.Waypoints
        for the Ship's current System

        .chart, .scan, and .survey are monkey-patched to have the
        ship already loaded as the first argument

        Blocks:
            False

        Returns:
            snisp.waypoints.Waypoints: Waypoints for the Ship's Location
        """
        waypoints = Waypoints(self.agent, self.location)
        waypoints.chart = functools.partial(waypoints.chart, self)
        waypoints.scan = functools.partial(waypoints.scan, self)
        waypoints.survey = functools.partial(waypoints.survey, self)
        return waypoints

    @property
    def markets(self):
        """
        Property that returns the snisp.markets.Markets
        for the Ship's current System

        Blocks:
            False

        Returns:
            snisp.markets.Markets: Markets for the Ship's Location
        """
        return Markets(self, self.location)

    @property
    def shipyards(self):
        """
        Property that returns the snisp.shipyards.Shipyards
        for the Ship's current System

        Blocks:
            False

        Returns:
            snisp.shipyards.Shipyards: Shipyards for the Ship's Location
        """
        return Shipyards(self, self.location)

    @property
    def system(self):
        """
        Property that returns the snisp.system.System
        for the Ship's current System

        .scan will be monkey-patched  to have the ship already
        loaded as the first argument

        Blocks:
            False

        Returns:
            snisp.systems.System: System for the Ship's Location
        """
        _system = System(self.agent)
        _system.scan = functools.partial(_system.scan, self)
        return _system

    @property
    def at_market(self):
        """
        Property that returns True if the Ship is DOCKED or IN_ORBIT
        at a Market; else False

        Blocks:
            False

        Returns:
            bool
        """
        if self.nav.status != 'IN_TRANSIT':
            waypoint = self.waypoints.get()
            return any(
                i.symbol.upper() == 'MARKETPLACE' for i in waypoint.traits
            )
        return False

    @property
    def at_shipyard(self):
        """
        Property that returns True if the Ship is DOCKED or IN_ORBIT
        at a Shipyard; else False

        Blocks:
            False

        Returns:
                bool
        """
        if self.nav.status != 'IN_TRANSIT':
            waypoint = self.waypoints.get()
            return any(
                i.symbol.upper() == 'SHIPYARD' for i in waypoint.traits
            )
        return False

    @property
    def can_mine(self):
        """
        Property that returns True if the Ship has a Mining Mount installed
        and can mine; else, False

        Blocks:
            False

        Returns:
                bool
        """
        for mount in self.mounts:
            if mount.symbol.upper().startswith('MOUNT_MINING_LASER_'):
                return True
        return False

    @property
    def can_refine_gas(self):
        """
        Property that returns True if the Ship has a Gas Processor
        Module installed and refine Gas; else, False

        Blocks:
            False

        Returns:
                bool
        """
        for module in self.modules:
            if module.symbol.upper().startswith('MODULE_GAS_PROCESSOR_'):
                return True
        return False

    @property
    def can_refine_ore(self):
        """
        Property that returns True if the Ship has a Mineral Processor
        Module installed and can refine Ore; else, False

        Blocks:
            False

        Returns:
                bool
        """
        for module in self.modules:
            if module.symbol.upper().startswith('MODULE_MINERAL_PROCESSOR_'):
                return True
        return False

    @property
    def can_siphon(self):
        """
        Property that returns True if the Ship has a Gas Siphon
        Mount installed and can siphon Gas; else, False

        Blocks:
            False

        Returns:
                bool
        """
        for mount in self.mounts:
            if mount.symbol.upper().startswith('MOUNT_GAS_SIPHON_'):
                return True
        return False

    @property
    def can_survey(self):
        """
        Property that returns True if the Ship has a Surveyor
        Mount installed and can survey; else, False

        Blocks:
            False

        Returns:
                bool
        """
        for mount in self.mounts:
            if mount.symbol.upper().startswith('MOUNT_SURVEYOR_'):
                return True
        return False

    def arrived_at_destination(self):
        current_nav = self.nav.to_dict()
        current_nav['status'] = 'IN_ORBIT'
        self.update_data_item('nav', current_nav)

    def closest(self, *iterables):
        """
        Iterates over supplied *iterables and returns the closest Waypoint
        to the Ship's current location

        Args:
            *iterables: Waypoints or Waypoint subclass types

        Blocks:
            False

        Returns:
            Waypoint or Waypoint subclass if *iterables return at least one
            Waypoint or Waypoint subclass; else, None
        """
        try:
            return min(
                (i for i in itertools.chain(*iterables)), key=self.distance
            )
        except ValueError:
            return

    def farthest(self, *iterables):
        """
        Iterates over supplied *iterables and returns the farthest Waypoint
        to the Ship's current location

        Args:
            *iterables: Waypoints or Waypoint subclass types

        Blocks:
            False

        Returns:
            Waypoint or Waypoint subclass if *iterables return at least one
            Waypoint or Waypoint subclass; else, None
        """
        try:
            return max(
                (i for i in itertools.chain(*iterables)), key=self.distance
            )
        except ValueError:
            return

    def distance(self, destination):
        """Returns the distance between the Ship and Waypoint

        Args:
            destination: Waypoint or Waypoint subclass type

        Blocks:
            False

        Returns:
            int
        """
        return utils.calculate_distance(self, destination)

    def closest_fuel(self):
        """
        Returns the closest Fuel Station to the Ship's current location.

        This is a time-consuming operation. The Library will do it's best
        to cache the result and use the cache on all subsequent calls

        Blocks:
            False

        Returns:
            Waypoint or Waypoint subclass
        """
        return self.closest(self.markets.fuel_stations())

    @retry()
    @transit
    def dock(self):
        """
        Dock the Ship at the current Waypoint, if it is not already docked

        If the Ship is IN_TRANSIT, the method will block unti the Ship
        has reached the destination

        Blocks:
            True: Won't be executed until Ship reaches destination
        """
        if self.nav.status != 'DOCKED':
            response = self.agent.client.post(
                f'/my/ships/{self.symbol}/dock'
            )
            data = response.json()['data']
            self.update_data_item('nav', data['nav'])
            logger.info(
                f'{self.registration.role}: {self.symbol} | '
                f'Docked at {self.nav.waypoint_symbol}'
            )
            return data

    @retry()
    @transit
    @cooldown
    @in_orbit
    def extract(self):
        """
        Make an extraction at the current Waypoint.

        Exceptions will be raised if the Ship is at an invalid Waypoint
        type or if it does not have an Extraction Mount

        If the Ship is IN_TRANSIT, the method will block unti the Ship
        has reached the destination.

        An Extraction causes a cooldown period. The method will
        automatically block if the Ship is in a cooldown period.

        Blocks:
            True: Won't be executed until Ship reaches destination
                  and/or cooldown has passed

        Returns:
            Extraction
        """
        response = self.agent.client.post(
            f'/my/ships/{self.symbol}/extract'
        )
        data = response.json()['data']
        self.update_data_item('cargo', data['cargo'])
        self.update_data_item('cooldown', data['cooldown'])
        logger.info(
            f'{self.registration.role}: {self.symbol} | Extracted '
            f'{data["extraction"]["yield"]["units"]:,} units of '
            f'{data["extraction"]["yield"]["symbol"]}'
        )
        return Extraction(self.agent, data['extraction']['yield'])

    @retry()
    @transit
    @cooldown
    @in_orbit
    def extract_with_survey(self, survey):
        """
        Make an extraction at the current Waypoint with the supplied Survey

        Exceptions will be raised if the Ship is at an invalid Waypoint
        type, if it does not have an Extraction Mount, or if the Survey
        is invalid

        If the Ship is IN_TRANSIT, the method will block unti the Ship
        has reached the destination.

        An Extraction causes a cooldown period. The method will
        automatically block if the Ship is in a cooldown period.

        Args:
            survey: A Survey from a self.survey().surveys

        Blocks:
            True: Won't be executed until Ship reaches destination
                  and/or cooldown has passed

        Returns:
            ExtractionWithSurvey
        """
        try:
            survey = survey.to_dict()
        except AttributeError:
            raise exceptions.ShipSurveyVerificationError(
                'extract_with_survey requires a Survey. '
                f'Received {type(survey)} instead'
            )
        response = self.agent.client.post(
            f'/my/ships/{self.symbol}/extract/survey', json=survey
        )
        data = response.json()['data']
        self.update_data_item('cargo', data['cargo'])
        self.update_data_item('cooldown', data['cooldown'])
        logger.info(
            f'{self.registration.role}: {self.symbol} | Extracted '
            f'{data["extraction"]["yield"]["units"]:,} units of '
            f'{data["extraction"]["yield"]["symbol"]}'
        )
        return ExtractionWithSurvey(self.agent, data['extraction']['yield'])

    @retry()
    @transit
    @docked
    def install_mount(self, mount_symbol):
        """
        Not currently implemented by SpaceTraders

        Installs the listed Mount onto the ship. Must be at a shipyard

        Args:
            mount_symbol: Symbol of the Mount to install.
                          See snisp.utils.SHIP_MOUNTS for all supported
                          Ship Mounts

        Blocks:
            True: Won't be executed until Ship reaches destination

        Returns:
           Mounts: The Ship's current Mounts
        """
        mount_symbol = mount_symbol.upper()
        if mount_symbol not in utils.SHIP_MOUNTS:
            raise exceptions.SpaceAttributeError(
                f'{mount_symbol} is not an acceptable mount type. '
                'See snisp.utils.SHIP_MOUNTS for acceptable mounts.'
            )
        with self.agent.lock:
            response = self.agent.client.post(
                f'/my/ships/{self.symbol}/mounts/install',
                json={'symbol': mount_symbol}
            )
            data = response.json()['data']
            self.update_data_item('cargo', data['cargo'])
            self.update_data_item('mounts', data['mounts'])
            logger.info(
                f'{self.registration.role}: {self.symbol} | '
                f'Installed mount {data["transaction"]["tradeSymbol"]} '
                f'for ${data["transaction"]["totalPrice"]:,.2f}'
            )
            transaction = Transaction(self.agent, data['transaction'])
            self.agent.recent_transactions.appendleft(transaction)
            return Mounts(self.agent, data['mounts'])

    @retry()
    def jettison(self, symbol, units=1):
        """
        Jettison the number of units for the symbol from the Ship's
        cargo.

        Args:
            symbol: The symbol of the item to jettison

        Kwargs:
            units: The number of units to jettison. Default is 1.
        """
        response = self.agent.client.post(
            f'/my/ships/{self.symbol}/jettison',
            json={'symbol': symbol, 'units': units}
        )
        data = response.json()['data']
        self.update_data_item('cargo', data['cargo'])
        logger.info(
            f'{self.registration.role}: {self.symbol} | '
            f'Jettisoning {units:,} of {symbol}'
        )
        return data

    @retry()
    @transit
    @cooldown
    @in_orbit
    def jump(self, waypoint):
        """
        Jump Ship to Waypoint. Requires the Ship to be at a full constructed
        JumpGate

        Blocks:
            True: Won't be executed until Ship reaches destination
                  and/or cooldown has passed

        Args:
            waypoint: Waypoint or Waypoint subclass type
        """
        response = self.agent.client.post(
            f'/my/ships/{self.symbol}/jump',
            json={'waypointSymbol': waypoint.symbol}
        )
        data = response.json()['data']
        self.update_data_item('nav', data['nav'])
        self.update_data_item('cooldown', data['cooldown'])
        logger.info(
            f'{self.registration.role}: {self.symbol} | '
            f'Jumped to {waypoint.symbol} via '
            f'Will reach destination in {self.arrival / 60:.1f} minutes'
        )
        return data

    def autopilot(
        self,
        waypoint,
        flight_mode='CRUISE',
        done_callback=None,
    ):
        """
        Navigates the Ship to the destination Waypoint.

        The method will block until the Ship has reached the destination.

        The method will automatically attempt to refuel and modify the Ship's
        flight_mode as needed.

        done_callback accepteds a callback function that will be performed
        before returning control to the thread

        Failed attempts to reach intermediate Waypoints or refuel attempts
        will be logged at level INFO

        Ships that cannot reach their destinations will be added to the Ship's
        agent.dead_ships dictionary. Dead Ships will not be returned on further
        agent.fleet iterations

        Args:
            waypoint: Waypoint or Waypoint subclass type

        Kwargs:
            flight_mode: Starting flight_mode. Default is CRUISE
            done_callback: callable() item that will be executed before
                           returning

        Blocks:
            True: until Ship reaches destination

        Returns:
            NavigateInsufficientFuelError: Ran out of Fuel
        """
        # A complete mess
        if self.nav.waypoint_symbol == waypoint.symbol:
            return
        current_nav = self.nav.to_dict()
        current_nav['status'] = 'IN_TRANSIT'
        self.update_data_item('nav', current_nav)
        self.update_flight_mode(flight_mode)
        self.refuel(ignore_errors=True)
        self.orbit()

        # Can make it to the dest as is. No need to do anything else
        if not self.navigate(waypoint, raise_error=False):
            self.refuel(ignore_errors=True)
            self.orbit()
            if callable(done_callback):
                done_callback()
            return

        closest_fuel_to_wp = sorted(
            self.markets.fuel_stations(),
            key=lambda x: utils.calculate_waypoint_distance(waypoint, x)
        )
        last_successful_fueling = len(closest_fuel_to_wp) + 1
        while self.navigate(waypoint, raise_error=False):
            fuel_index = 0
            next_fuel = closest_fuel_to_wp[fuel_index]
            if self.nav.waypoint_symbol == next_fuel.symbol:
                self.refuel(ignore_errors=True)
                if self.nav.flight_mode == 'DRIFT':
                    with self.agent.lock:
                        self.agent.dead_ships[self.symbol] = self
                    raise exceptions.NavigateInsufficientFuelError(
                        r'¯\_(ツ)_/¯'
                    )
                logger.info(
                    f'{self.registration.role}: {self.symbol} | '
                    'Already at closest fuel '
                    f'but cannot make it to {waypoint.symbol} at '
                    f'{self.nav.flight_mode}. Will now DRIFT.'
                )
                self.update_flight_mode('DRIFT')
                continue
            while self.navigate(next_fuel, raise_error=False):
                fuel_index += 1
                if fuel_index >= last_successful_fueling:
                    # There's a blackhole where the ship could
                    # get to the closest fueling station to the waypoint
                    # but not make it to the waypoint from there.
                    # It would just continually cycle fuel stations if
                    # the flight_mode was not switched to drift.
                    # If it's *already* as DRIFT at this point, well, :/
                    if self.nav.flight_mode == 'DRIFT':
                        if err := self.navigate(waypoint, raise_error=False):
                            try:
                                logger.warning(err.data)
                            except Exception as e:
                                logger.warning(e)
                            with self.agent.lock:
                                self.agent.dead_ships[self.symbol] = self
                            raise exceptions.NavigateInsufficientFuelError(
                                'Waypoint too far away. SOL'
                            )
                        else:
                            self.orbit()
                            if callable(done_callback):
                                done_callback()
                            return
                    else:
                        logger.info(
                            f'{self.registration.role}: {self.symbol} | '
                            'At last successful fuel stop.  Will now DRIFT.'
                        )
                        self.update_flight_mode('DRIFT')
                        break
                try:
                    next_fuel = closest_fuel_to_wp[fuel_index]
                    logger.info(
                        f'{self.registration.role}: {self.symbol} | '
                        f'Trying next FUEL_STATION: {next_fuel.symbol} | '
                        f'{fuel_index}/{len(closest_fuel_to_wp)}'
                    )
                except IndexError:
                    if self.nav.flight_mode == 'DRIFT':
                        if err := self.navigate(waypoint, raise_error=False):
                            try:
                                logger.warning(err.data)
                            except Exception as e:
                                logger.warning(e)
                            with self.agent.lock:
                                self.agent.dead_ships[self.symbol] = self
                            raise exceptions.NavigateInsufficientFuelError(
                                'No close fuel. SOL, buddy'
                            )
                        else:
                            self.orbit()
                            if callable(done_callback):
                                done_callback()
                            return
                    if (self.fuel.current + 5) >= self.fuel.capacity:
                        # Can't make it to the WP or any close fuel
                        # Weird situation where it'll make the drift
                        # to the closest_fuel_to_wp but also the farthest...
                        self.update_flight_mode('DRIFT')
                    else:
                        # May have started with less than a full fuel tank
                        logger.warning(
                            f'{self.registration.role}: {self.symbol} | '
                            'No CRUISABLE fuel stations. Now drifting.'
                        )
                        self.update_flight_mode('DRIFT')
                        closest_drift_fuel = self.closest_fuel()
                        if err := self.navigate(
                            closest_drift_fuel, raise_error=False
                        ):
                            try:
                                logger.warning(err.data)
                            except Exception as e:
                                logger.warning(e)
                            with self.agent.lock:
                                self.agent.dead_ships[self.symbol] = self
                            raise exceptions.NavigateInsufficientFuelError(
                                'No close fuel. ya burnt'
                            )
                        self.refuel()
                        self.update_flight_mode('CRUISE')
                    break
            else:
                last_successful_fueling = fuel_index
            self.refuel(ignore_errors=True)
        self.refuel(ignore_errors=True)
        self.orbit()
        if callable(done_callback):
            done_callback()

    @retry()
    @transit
    @in_orbit
    def navigate(self, waypoint, raise_error=True):
        """
        Attempts to navigate the Ship to the Waypoint.

        If the Ship is already IN_TRANSIT, the method will block
        until it reaches the previous destination first.

        Optional raise_error kwarg can suppress or raise Exceptions.
        Default is True

        Args:
            waypoint: Waypoint or Waypoint subclass type

        Kwargs:
            raise_error: If False, errors will be suppressed and returned.
                         Default True

        Blocks:
            True: Will block if Ship is already IN_TRANSIT. Oherwise, False

        Returns:
            Exception: Returns an Exception if one is raised and rase_error
                       is False
        """
        if self.nav.waypoint_symbol == waypoint.symbol:
            return
        try:
            response = self.agent.client.post(
                f'/my/ships/{self.symbol}/navigate',
                json={'waypointSymbol': waypoint.symbol}
            )
        except Exception as e:
            if raise_error:
                raise e
            else:
                return e
        data = response.json()['data']
        self.update_data_item('fuel', data['fuel'])
        self.update_data_item('nav', data['nav'])
        logger.info(
            f'{self.registration.role}: {self.symbol} | '
            f'Navigating to {waypoint.symbol} via '
            f'{self.nav.flight_mode}. Will reach destination in '
            f'{self.arrival / 60:.1f} minutes'
        )

    @retry()
    @transit
    @docked
    def negotiate_contract(self):
        """
        Negotiates a new Contract.

        The Ship must be at a Wayoint with a Faction offering a Contract
        and the Agent associated with the Ship must not already have an active
        Contract

        Blocks:
            True: Won't be executed until Ship reaches destination

        Returns:
            Contract
        """
        response = self.agent.client.post(
            f'/my/ships/{self.symbol}/negotiate/contract'
        )
        logger.info(
            f'{self.registration.role}: {self.symbol} | '
            'Negotiated new contract'
        )
        return Contract(self.agent, response.json()['data']['contract'])

    @retry()
    @transit
    def orbit(self):
        """
        Orbit the Ship at the current Waypoint, if it is not already inorbit

        Blocks:
            True: Won't be executed until Ship reaches destination
        """
        if self.nav.status != 'IN_ORBIT':
            response = self.agent.client.post(
                f'/my/ships/{self.symbol}/orbit'
            )
            data = response.json()['data']
            self.update_data_item('nav', data['nav'])
            logger.info(
                f'{self.registration.role}: {self.symbol} | Moved to orbit'
            )
            return data

    @transit
    def autopurchase(self, trade_symbol, max_units=0, buffer=200_000):
        """
        Purchases up to max_units of trade_symbol, depending on the Ship's
        cargo capacity and the Agent's current credits with regards to
        the buffer

        If max_units <= 0, max_units will be up to how many units
        are available in the Ship's cargo; else, max_units will be up to the
        number supplied

        Buffer will be your credits buffer. The default 200_000 limit means you
        will be able to purchase up to max_units so long as your current
        agent.data.credits - buffer >= purchase price.
        To remove the buffer, just pass a 0

        Args:
            trade_symbol: Symbol of item to purchase.

        Kwargs:
            max_units: Up to max_units to purchase. Default is until the ship's
                       cargo is at capacity
            buffer: Credit buffer. Up to max_units will be purchased so far as
                    the purchase amount - agent.data.credits - buffer > 0.
                    Default is 200_000

        Blocks:
            True: Won't be executed until Ship reaches destination

        Returns:
            Transactions:  List of successful Transactions or an empty list
        """
        trade_symbol = trade_symbol.upper()
        if trade_symbol not in utils.GOODS_TYPES:
            raise exceptions.SpaceAttributeError(
                f'{trade_symbol} is not an acceptable goods type. '
                'See snisp.utils.GOODS_TYPES for acceptable goods.'
            )
        max_units = int(max_units)
        if max_units <= 0:
            max_units = self.cargo.capacity
        transactions = []
        market_data = self.markets()
        if not market_data:
            logger.warning(
                f'{self.registration.role}: {self.symbol} | '
                f'No market at {self.location.waypoint} '
                f'Cannot purchase {max_units:,} of {trade_symbol}.'
            )
            return transactions
        trade_volume = None
        for t_good in market_data.trade_goods:
            if t_good.symbol == trade_symbol:
                trade_volume = t_good.trade_volume
                break
        if trade_volume is None:
            logger.warning(
                f'Market {self.location.waypoint} no '
                f'longer trading {trade_symbol}'
            )
            return transactions
        with self.agent.lock:
            while max_units > 0:
                unit_price = None
                for t_good in self.markets().trade_goods:
                    if t_good.symbol == trade_symbol:
                        unit_price = t_good.purchase_price
                        break
                if unit_price is None:
                    logger.warning(
                        f'Market {self.location.waypoint} no longer '
                        f'trading {trade_symbol}'
                    )
                    return transactions
                buy_units = max_units
                if buy_units > trade_volume:
                    buy_units = trade_volume
                buy_units = min(
                    buy_units,
                    self.cargo.capacity - self.cargo.units
                )
                if buy_units <= 0:
                    return transactions
                purchase_amount = buy_units * unit_price
                cash = self.agent.data.credits
                while purchase_amount > cash - buffer:
                    buy_units -= 1
                    if buy_units == 0:
                        return transactions
                    purchase_amount = buy_units * unit_price
                transactions.append(self.purchase(trade_symbol, buy_units))
                max_units -= buy_units
        return transactions

    @retry()
    @transit
    @docked
    def purchase(self, trade_symbol, units):
        """
        Purchases the specified units of the trade_symbol

        Examples of exceptions are insufficient funds, cargo capacity,
        or if at an invalid waypoint

        Args:
            trade_symbol: Symbol of item to purchase
            units: Number of units to purchase

        Blocks:
            True: Won't be executed until Ship reaches destination

        Returns:
            Transacaction
        """
        trade_symbol = trade_symbol.upper()
        if trade_symbol not in utils.GOODS_TYPES:
            raise exceptions.SpaceAttributeError(
                f'{trade_symbol} is not an acceptable goods type. '
                'See snisp.utils.GOODS_TYPES for acceptable goods.'
            )
        with self.agent.lock:
            payload = {'symbol': trade_symbol, 'units': int(units)}
            response = self.agent.client.post(
                f'/my/ships/{self.symbol}/purchase', json=payload
            )
            data = response.json()['data']
            self.update_data_item('cargo', data['cargo'])
            logger.info(
                f'{self.registration.role}: {self.symbol} | Purchased '
                f'{data["transaction"]["units"]:,}'
                f' units of {data["transaction"]["tradeSymbol"]} for '
                f'${data["transaction"]["totalPrice"]:,.2f}'
            )
            transaction = Transaction(self.agent, data['transaction'])
            self.agent.recent_transactions.appendleft(transaction)
            return transaction

    @retry()
    @transit
    @docked
    def refuel(self, *, units=0, from_cargo=False, ignore_errors=False):
        """
        Refuels the Ship if the Ship is at a Waypoint that sells Fuel, or,
        if from_cargo is True, Refuels from the Ship cargo

        Kwargs:
            units: If greather than 0, restrict to N units being refilled.
                   Default 0 will refill the fuel tank completely
            from_cargo: If True, refuel from Cargo. Default False
            ignore_errors: If True, errors will be suppressed and returned.
                           Default False

        Blocks:
            True: Won't be executed until Ship reaches destination

        Returns:
            Exception: Returns an Exception if one occurs and
                       ignore_errors is True
        """
        if self.fuel.capacity == self.fuel.current:
            return
        payload = {'fromCargo': from_cargo}
        if (units := int(units)) >= 1:
            payload['units'] = units
        with self.agent.lock:
            try:
                response = self.agent.client.post(
                    f'/my/ships/{self.symbol}/refuel', json=payload
                )
            except Exception as e:
                if ignore_errors:
                    logger.warning(
                        f'{self.registration.role}: {self.symbol} | '
                        f'Failed to refuel at {self.location.waypoint}.'
                    )
                    return
                raise e
            data = response.json()['data']
            self.update_data_item('fuel', data['fuel'])
            if from_cargo:
                cargo = self.to_dict()['cargo']
                units = data['transaction']['units']
                for index, item in enumerate(cargo['inventory']):
                    if item['symbol'].upper() == 'FUEL':
                        item['units'] -= units
                        if item['units'] == 0:
                            del cargo['inventory'][index]
                self.update_data_item('cargo', cargo)
            logger.info(
                f'{self.registration.role}: {self.symbol} | Refueled '
                f'{data["transaction"]["units"]:,} '
                f'units at {data["transaction"]["waypointSymbol"]} for '
                f'${data["transaction"]["totalPrice"]:,.2f}'
            )
            return Transaction(self.agent, data['transaction'])

    @retry()
    @cooldown
    def refine(self, produce):
        """Refine Ore into a product

        Converts 30 units of Ore into 1 inut of Product

        Args:
            produce: The symbol to produce. i.e., "IRON".
                     See snisp.utils.REFINABLE_SYMBOLS for refinable types

        Blocks:
            True: Blocks during cooldown
        """
        if produce not in utils.REFINABLE_SYMBOLS:
            raise exceptions.ShipInvalidRefineryGoodError(
                f'{produce} is not an acceptable produce. '
                'See snisp.utils.REFINABLE_SYMBOLS for acceptable produces.'
            )
        response = self.agent.client.post(
            f'/my/ships/{self.symbol}/refine', json={'produce': produce}
        )
        data = response.json()['data']
        self.update_data_item('cargo', data['cargo'])
        self.update_data_item('cooldown', data['cooldown'])
        logger.info(
            f'{self.registration.role}: {self.symbol} | Refined '
            f'{data["produced"][0]["units"]:,} units of'
            f'{data["produced"][0]["tradeSymbol"]}'
        )
        return data

    @retry()
    def refresh(self):
        """Returns a new Ship class object of the current ship

        Returns:
            A new Ship instance from self
        """
        response = self.agent.client.get(f'/my/ships/{self.symbol}')
        return Ship(self.agent, response.json()['data'])

    @retry()
    @transit
    @docked
    def remove_mount(self, mount_symbol):
        """
        Not currently implemented by SpaceTraders

        Removes the listed Mount onto the ship. Must be at a Shipyard

        Args:
            mount_symbol: Symbol of the Mount to remove

        Blocks:
            True: Won't be executed until Ship reaches destination

        Returns:
           Mounts: The Ship's current Mounts
        """
        response = self.agent.client.post(
            f'/my/ships/{self.symbol}/mounts/remove',
            json={'symbol': mount_symbol}
        )
        data = response.json()['data']
        self.update_data_item('cargo', data['cargo'])
        self.update_data_item('mounts', data['mounts'])
        logger.info(
            f'{self.registration.role}: {self.symbol} | '
            'Removed mount {mount_symbol}'
        )
        return data

    @retry()
    @transit
    @docked
    def repair_cost(self):
        """
        Property that returns the current cost to repair the ship

        Ship must be located at a Shipyard

        Blocks:
            True: Won't be executed until Ship reaches destination

        Returns:
            int
        """
        response = self.agent.client.get(f'/my/ships/{self.symbol}/repair')
        data = response.json()['data']
        return int(data['transaction']['totalPrice'])

    @retry()
    @transit
    @docked
    def repair(self):
        """
        Repairs the ship to full working condition

        Ship must be located at a Shipyard

        Blocks:
            True: Won't be executed until Ship reaches destination

        Returns:
            Transaction
        """
        payload = {'shipSymbol': self.symbol}
        response = self.agent.client.post(
            f'/my/ships/{self.symbol}/repair', json=payload
        )
        data = response.json()['data']
        self.update_data_item('frame', data['ship']['frame'])
        self.update_data_item('reactor', data['ship']['reactor'])
        self.update_data_item('engine', data['ship']['engine'])
        logger.info(
            f'{self.registration.role}: {self.symbol} | '
            f'Repaired ship for ${data["transaction"]["totalPrice"]:,.2f}'
        )
        transaction = Transaction(self.agent, data['transaction'])
        self.agent.recent_transactions.appendleft(transaction)
        return transaction

    @retry()
    @transit
    @in_orbit
    @cooldown
    def scan(self):
        """
        Scan current System for ships

        NOTE: SpaceTraders does not make it explicit who is the owner of a
              scanned ship. SnakesInSpace currently assumes scanned
              ships are owned by the Agent. Attempting to perform actions on
              a ship you do not own will lead to undefined consequences.

        Blocks:
            True: Won't be executed until Ship reaches destination
                  and/or cooldown has passed

        Yields:
            Ships
        """
        response = self.agent.client.post(
            f'/my/ships/{self.symbol}/scan/ships'
        )
        data = response.json()['data']
        self.update_data_item('cooldown', data['cooldown'])
        for ship in data['ships']:
            yield Ship(self.agent, ship)

    @retry()
    @transit
    @docked
    def scrap_price(self):
        """
        Property that returns the current scrap price for the ship

        Ship must be located at a Shipyard

        Blocks:
            True: Won't be executed until Ship reaches destination

        Returns:
            int
        """
        response = self.agent.client.get(f'/my/ships/{self.symbol}/scrap')
        data = response.json()['data']
        return int(data['transaction']['totalPrice'])

    @retry()
    @transit
    @docked
    def scrap(self):
        """
        Sells the ship as scrap to the Shipyard at the ship's location.

        Ship must be located at a Shipyard

        Blocks:
            True: Won't be executed until Ship reaches destination

        Returns:
            Transaction
        """
        payload = {'shipSymbol': self.symbol}
        response = self.agent.client.post(
            f'/my/ships/{self.symbol}/scrap', json=payload
        )
        data = response.json()['data']
        logger.info(
            f'{self.registration.role}: {self.symbol} | '
            f'Scraped ship for ${data["transaction"]["totalPrice"]:,.2f}'
        )
        transaction = Transaction(self.agent, data['transaction'])
        self.agent.recent_transactions.appendleft(transaction)
        return transaction

    @transit
    def sell_all(self, trade_symbol, units=0):
        """
        Sell goods from the Ship's cargo.

        Args:
            trade_symbol: Good type. Symbol of item to sell

        Kwargs:
            units: The number of units to sell. If units < 0, all will be sold
                   Default 0

        Blocks:
            True: Won't be executed until Ship reaches destination

        Returns:
            Transactions:  List of successful Transactions or an empty list
        """
        transactions = []
        units = int(units)
        if units <= 0:
            units = next(
                (
                    i.units
                    for i in self.cargo.inventory if i.symbol == trade_symbol
                ), None
            )
            if units is None:
                logger.warning(
                    f'{self.registration.role}: {self.symbol} | '
                    'Attempting to sell {trade_symbol} which '
                    'is not currently in its cargo hold'
                )
                return transactions
        market_data = self.markets()
        if not market_data:
            logger.warning(
                f'{self.registration.role}: {self.symbol} | '
                f'No market at {self.location.waypoint}. '
                f'Cannot sell {units:,} of {trade_symbol}.'
            )
            return transactions
        trade_volume = next(
            (
                i.trade_volume
                for i in market_data.trade_goods
                if i.symbol == trade_symbol
            ), None
        )
        if trade_volume is None:
            logger.warning(
                f'Market {market_data.symbol} no longer trading {trade_symbol}'
            )
            return transactions
        while units:
            sell_units = units
            if units > trade_volume:
                units -= trade_volume
                sell_units = trade_volume
            else:
                units = 0
            transactions.append(self.sell(trade_symbol, sell_units))
        self.agent.recent_transactions.extendleft(transactions)
        return transactions

    def sell_off_cargo(self, trade_symbol=None):
        """
        Sell all goods from the Ship's cargo. Will automatically navigate
        to closest Market which sells the respective goods

        Kwargs:
            trade_symbol: Good type. Symbol of item to sell otherwise sells
                          everything in ship.cargo.inventory. Default is None

        Blocks:
            True: Won't be executed until Ship reaches destination

        Returns:
            Transactions:  List of successful Transactions or an empty list
        """

        transactions = []
        for item in self.cargo.inventory:
            if trade_symbol and item.symbol != trade_symbol:
                continue
            markets = (i for i in self.markets.imports(item.symbol))
            if market := self.closest(markets):
                self.autopilot(market)
                last_transactions = self.sell_all(item.symbol, item.units)
                if not last_transactions:
                    self.jettison(item.symbol, item.units)
                else:
                    transactions.extend(last_transactions)
            else:
                logger.info(
                    f'{self.registration.role}: {self.symbol} | '
                    f'No market buys {item.symbol}'
                )
                self.jettison(item.symbol, item.units)
        return transactions

    @retry()
    @transit
    @docked
    def sell(self, trade_symbol, units):
        """
        Sell goods from the Ship's cargo.

        Args:
            trade_symbol: Good type. Symbol of item to sell
            units: The number of units to sell

        Blocks:
            True: Won't be executed until Ship reaches destination

        Returns:
            Transaction
        """
        trade_symbol = trade_symbol.upper()
        if trade_symbol not in utils.GOODS_TYPES:
            raise exceptions.SpaceAttributeError(
                f'{trade_symbol} is not an acceptable goods type. '
                'See snisp.utils.GOODS_TYPES for acceptable goods.'
            )
        payload = {'symbol': trade_symbol, 'units': int(units)}
        with self.agent.lock:
            response = self.agent.client.post(
                f'/my/ships/{self.symbol}/sell', json=payload
            )
            data = response.json()['data']
            self.update_data_item('cargo', data['cargo'])
            logger.info(
                f'{self.registration.role}: {self.symbol} | '
                f'Sold {data["transaction"]["units"]:,} '
                f'units of {data["transaction"]["tradeSymbol"]} for '
                f'${data["transaction"]["totalPrice"]:,.2f}'
            )
            transaction = Transaction(self.agent, data['transaction'])
            self.agent.recent_transactions.appendleft(transaction)
            return transaction

    @retry()
    @transit
    @cooldown
    @in_orbit
    def siphon(self):
        """
        Make a siphon at the current Waypoint.

        Exceptions will be raised if the Ship is at an invalid Waypoint
        type or if it does not have a Siphon Mount

        A Siphon causes a cooldown period. The method will
        automatically block if the Ship is in a cooldown period.

        Blocks:
            True: Won't be executed until Ship reaches destination
                  and/or cooldown has passed

        Returns:
            Siphon
        """
        response = self.agent.client.post(
            f'/my/ships/{self.symbol}/siphon'
        )
        data = response.json()['data']
        self.update_data_item('cargo', data['cargo'])
        self.update_data_item('cooldown', data['cooldown'])
        logger.info(
            f'{self.registration.role}: {self.symbol} | Siphoned '
            f'{data["siphon"]["yield"]["units"]:,} units of '
            f'{data["siphon"]["yield"]["symbol"]}'
        )
        return Siphon(self.agent, data['siphon']['yield'])

    @retry()
    @transit
    def transfer(self, receive_ship, *, symbol, units):
        """
        Transfer cargo from Ship to another Ship

        Updates both Ships' cargos. Requires both Ships to be at the
        same Waypoint

        Args:
            receive_ship: The snisp.fleet.Ship instance of another Ship

        Kwargs:
            symbol: The symbol for the goods to transfer
            units: The number of units to transfer

        Blocks:
            True: until both Ships reach destination
        """
        symbol = symbol.upper()
        if symbol not in utils.GOODS_TYPES:
            raise exceptions.SpaceAttributeError(
                f'{symbol} is not an acceptable goods type. '
                'See snisp.utils.GOODS_TYPES for acceptable goods.'
            )
        units = int(units)
        payload = {
            'tradeSymbol': symbol,
            'units': units,
            'shipSymbol': receive_ship.symbol,
        }
        response = self.agent.client.post(
            f'/my/ships/{self.symbol}/transfer', json=payload
        )
        # NOTE: The response body's cargo shows the cargo of the
        #       transferring ship after the transfer is complete.
        data = response.json()['data']
        self.update_data_item('cargo', data['cargo'])
        new_cargo = receive_ship.cargo.to_dict()
        if not new_cargo.get('inventory'):
            # Placeholder name and description
            new_cargo['inventory'] = [
                {
                    'symbol': symbol,
                    'units': units,
                    'name': symbol,
                    'description': symbol,
                }
            ]
        else:
            if exists := next(
                (
                    i for i in new_cargo['inventory']
                    if i['symbol'] == symbol
                ), None
            ):
                exists['units'] += units
                new_cargo['units'] += units
            else:
                new_cargo['inventory'].append(
                    {
                        'symbol': symbol,
                        'units': units,
                        'name': symbol,
                        'description': symbol,
                    }
                )
                new_cargo['units'] += units
        receive_ship.update_data_item('cargo', new_cargo)
        logger.info(
            f'{self.registration.role}: {self.symbol} | '
            f'Transferred {units:,} of {symbol} '
            f'to {receive_ship.symbol}'
        )
        return data

    @retry()
    @transit
    @in_orbit
    def warp(self, waypoint):
        """
        Warp Ship to Waypoint.

        Requires the Ship to have a Warp Drive

        Args:
            waypoint: Waypoint or Waypoint subclass type

        Blocks:
            True: Won't be executed until Ship reaches destination
        """
        response = self.agent.client.post(
            f'/my/ships/{self.symbol}/warp',
            json={'waypointSymbol': waypoint.symbol}
        )
        data = response.json()['data']
        self.update_data_item('nav', data['nav'])
        self.update_data_item('fuel', data['fuel'])
        logger.info(
            f'{self.registration.role}: {self.symbol} | '
            f'Warping to {waypoint.symbol} '
            f'Will reach destination in {self.arrival / 60:.1f} minutes'
        )
        return data

    @retry()
    @transit
    def update_flight_mode(self, mode):
        """
        Updates the Ships flight mode if needed

        Args:
            mode: New flight_mode. See snips.utils.FLIGHT_MODES for supported
                  flight modes

        Blocks:
            True: Won't be executed until Ship reaches destination
        """
        mode = mode.strip().upper()
        if mode not in utils.FLIGHT_MODES:
            raise exceptions.SpaceAttributeError(
                f'{mode} is not an acceptable flight mode type. '
                'See snisp.utils.FLIGHT_MODES for acceptable flight modes.'
            )
        if self.nav.flight_mode != mode:
            prev_mode = self.nav.flight_mode
            response = self.agent.client.patch(
                f'/my/ships/{self.symbol}/nav', json={'flightMode': mode}
            )
            data = response.json()['data']
            self.update_data_item('nav', data)
            logger.info(
                f'{self.registration.role}: {self.symbol} | '
                f'Changing FLIGHT_MODE from {prev_mode} to {mode}'
            )
            return data


class Extraction(utils.AbstractJSONItem):

    def __init__(self, agent, extraction):
        self.agent = agent
        self._data = extraction


class ExtractionWithSurvey(utils.AbstractJSONItem):

    def __init__(self, agent, extraction):
        self.agent = agent
        self._data = extraction


class Siphon(utils.AbstractJSONItem):

    def __init__(self, agent, siphon):
        self.agent = agent
        self._data = siphon


class Mounts(utils.AbstractJSONItem):

    def __init__(self, agent, mount):
        self.agent = agent
        self._data = mount


class Transaction(utils.AbstractJSONItem):

    def __init__(self, agent, transaction):
        self.agent = agent
        self._data = transaction



import httpx
import json
import os
import pytest

import snisp

from . import DATA_DIR


class TestFactions:

    agent = snisp.agent.Agent(symbol='testing', faction='testing')

    def test_repr(self, respx_mock):
        assert repr(self.agent.factions) == f'Factions({self.agent!r})'

    @pytest.mark.respx(base_url='https://api.spacetraders.io/v2')
    def test_call(self, respx_mock):
        faction_data = json.load(
            open(os.path.join(DATA_DIR, 'faction.json'), encoding='utf8')
        )
        for symbol in snisp.utils.FACTION_SYMBOL:
            faction_data['data']['symbol'] = symbol
            respx_mock.get(f'/factions/{symbol}').mock(
                return_value=httpx.Response(200, json=faction_data)
            )
            faction = self.agent.factions(symbol)
            assert faction.to_dict() == faction_data['data']

        with pytest.raises(snisp.exceptions.WaypointNoFactionError):
            self.agent.factions('INVALID')

    @pytest.mark.respx(base_url='https://api.spacetraders.io/v2')
    def test_iter(self, respx_mock):
        factions_data = json.load(
            open(os.path.join(DATA_DIR, 'factions.json'), encoding='utf8')
        )
        respx_mock.get(
            '/factions', params={'page': 1, 'limit': 20}
        ).mock(
            return_value=httpx.Response(200, json=factions_data)
        )
        respx_mock.get(
            '/factions', params={'page': 2, 'limit': 20}
        ).mock(
            return_value=httpx.Response(200, json={'data': []})
        )
        factions = list(self.agent.factions)
        assert len(factions) == 1
        assert factions[0].symbol == 'COSMIC'



import copy
import httpx
import inspect
import json
import os
import pytest

from datetime import datetime, timedelta, timezone
from respx.patterns import M

import snisp

from . import DATA_DIR, GenericSideEffect


class TestWaypoints:

    agent = snisp.agent.Agent(symbol='testing', faction='testing')

    def test_repr(self, respx_mock):
        assert repr(self.agent.factions) == f'Factions({self.agent!r})'

    @pytest.mark.respx(base_url='https://api.spacetraders.io/v2')
    def test_chart(self, respx_mock):
        ship_data = json.load(
            open(os.path.join(DATA_DIR, 'ship_info.json'), encoding='utf8')
        )
        respx_mock.get('/my/ships/TEST_SHIP_SYMBOL').mock(
            return_value=httpx.Response(200, json=ship_data)
        )
        chart_data = json.load(
            open(os.path.join(DATA_DIR, 'chart.json'), encoding='utf8')
        )
        chart_side_effect = ChartSideEffect(chart_data)
        chart_route = respx_mock.post('/my/ships/TEST_SHIP_SYMBOL/chart')
        chart_route.side_effect = chart_side_effect

        ship = self.agent.fleet('TEST_SHIP_SYMBOL')
        chart = ship.waypoints.chart()
        assert chart.to_dict() == chart_data['data']

        # 4230: WaypointChartedError
        with chart_side_effect as tmp:
            tmp.already_charted = True
            ship = self.agent.fleet('TEST_SHIP_SYMBOL')
            chart = ship.waypoints.chart()
            assert chart.to_dict() == {}

        # .raised_at
        with chart_side_effect as tmp:
            tmp.fail = True
            ship = self.agent.fleet('TEST_SHIP_SYMBOL')
            try:
                chart = ship.waypoints.chart()
            except snisp.exceptions.ClientError as e:
                _ = e.raised_at

    @pytest.mark.respx(base_url='https://api.spacetraders.io/v2')
    def test_scan(self, respx_mock):
        ship_data = json.load(
            open(os.path.join(DATA_DIR, 'ship_info.json'), encoding='utf8')
        )
        respx_mock.get('/my/ships/TEST_SHIP_SYMBOL').mock(
            return_value=httpx.Response(200, json=ship_data)
        )
        scan_data = json.load(
            open(
                os.path.join(DATA_DIR, 'scan_waypoints.json'), encoding='utf8'
            )
        )
        scan_side_effect = GenericSideEffect(scan_data)
        scan_route = respx_mock.post(
            '/my/ships/TEST_SHIP_SYMBOL/scan/waypoints'
        )
        scan_route.side_effect = scan_side_effect
        orbit_data = json.load(
            open(os.path.join(DATA_DIR, 'orbit.json'), encoding='utf8')
        )
        respx_mock.post('/my/ships/TEST_SHIP_SYMBOL/orbit').mock(
            return_value=httpx.Response(200, json=orbit_data)
        )

        ship = self.agent.fleet('TEST_SHIP_SYMBOL')
        scan = ship.waypoints.scan()
        assert inspect.isgenerator(scan)
        scans = list(scan)
        assert ship.cooldown.total_seconds == 60
        assert ship.cooldown.remaining_seconds == 59
        assert len(scans) == 1
        for scan, scan_json in zip(scans, scan_data['data']['waypoints']):
            assert scan.to_dict() == scan_json

    @pytest.mark.respx(base_url='https://api.spacetraders.io/v2')
    def test_best_survey(self, respx_mock):
        ship_data = json.load(
            open(os.path.join(DATA_DIR, 'ship_info.json'), encoding='utf8')
        )
        ship_data['data']['nav']['systemSymbol'] = 'TEST-SYSTEM'
        ship_data['data']['nav']['waypointSymbol'] = 'TEST-SYSTEM-WAYPOINT'
        ship_data['data']['nav']['route']['destination']['type'] = 'ASTEROID'
        ship_data['data']['nav']['status'] = 'IN_ORBIT'
        ship_side_effect = GenericSideEffect(ship_data)
        ship_route = respx_mock.get('/my/ships/TEST_SHIP_SYMBOL')
        ship_route.side_effect = ship_side_effect

        survey_data = json.load(
            open(os.path.join(DATA_DIR, 'survey.json'), encoding='utf8')
        )
        expires = str(datetime.now(timezone.utc) + timedelta(hours=1))
        for survey in survey_data['data']['surveys']:
            survey['expiration'] = expires
        survey_side_effect = SurveySideEffect(survey_data)
        survey_route = respx_mock.post('/my/ships/TEST_SHIP_SYMBOL/survey')
        survey_route.side_effect = survey_side_effect

        ship = self.agent.fleet('TEST_SHIP_SYMBOL')
        survey = ship.waypoints.survey()
        best = survey.best()
        assert best.size == 'LARGE'

        best = survey.best('INVALID')
        assert not best

        best = survey.best('PRECIOUS_STONES', 'IRON_ORE')
        assert best.size == 'LARGE'
        symbols = {i.symbol for i in best.deposits}
        assert 'IRON_ORE' in symbols

    @pytest.mark.respx(base_url='https://api.spacetraders.io/v2')
    def test_survey(self, respx_mock):
        ship_data = json.load(
            open(os.path.join(DATA_DIR, 'ship_info.json'), encoding='utf8')
        )
        ship_data['data']['nav']['systemSymbol'] = 'TEST-SYSTEM'
        ship_data['data']['nav']['waypointSymbol'] = 'TEST-SYSTEM-WAYPOINT'
        ship_data['data']['nav']['route']['destination']['type'] = 'ASTEROID'
        ship_data['data']['nav']['status'] = 'IN_ORBIT'
        ship_side_effect = GenericSideEffect(ship_data)
        ship_route = respx_mock.get('/my/ships/TEST_SHIP_SYMBOL')
        ship_route.side_effect = ship_side_effect

        survey_data = json.load(
            open(os.path.join(DATA_DIR, 'survey.json'), encoding='utf8')
        )
        expires = str(datetime.now(timezone.utc) + timedelta(hours=1))
        for survey in survey_data['data']['surveys']:
            survey['expiration'] = expires
        survey_side_effect = SurveySideEffect(survey_data)
        survey_route = respx_mock.post('/my/ships/TEST_SHIP_SYMBOL/survey')
        survey_route.side_effect = survey_side_effect

        ship = self.agent.fleet('TEST_SHIP_SYMBOL')
        survey = ship.waypoints.survey()
        assert len(survey.surveys) == 3
        assert ship.cooldown.total_seconds == 60
        assert ship.cooldown.remaining_seconds == 59

        # 4240: ShipMissingSurveyorError
        with survey_side_effect as tmp:
            tmp.no_mounts = True
            ship = self.agent.fleet('TEST_SHIP_SYMBOL')
            with pytest.raises(
                snisp.exceptions.ShipMissingSurveyorError
            ):
                ship.waypoints.survey()

        # 4222: ShipSurveyWaypointTypeError
        with ship_side_effect as tmp:
            tmp.data['data']['nav']['route']['destination']['type'] = 'PLANET'
            ship = self.agent.fleet('TEST_SHIP_SYMBOL')
            with pytest.raises(
                snisp.exceptions.ShipSurveyWaypointTypeError
            ):
                ship.waypoints.survey()

    @pytest.mark.respx(base_url='https://api.spacetraders.io/v2')
    def test_surveys(self, respx_mock):
        ship_data = json.load(
            open(os.path.join(DATA_DIR, 'ship_info.json'), encoding='utf8')
        )
        ship_data['data']['nav']['systemSymbol'] = 'TEST-SYSTEM'
        ship_data['data']['nav']['waypointSymbol'] = 'TEST-SYSTEM-WAYPOINT'
        ship_data['data']['nav']['route']['destination']['type'] = 'ASTEROID'
        ship_data['data']['nav']['status'] = 'IN_ORBIT'
        ship_side_effect = GenericSideEffect(ship_data)
        ship_route = respx_mock.get('/my/ships/TEST_SHIP_SYMBOL')
        ship_route.side_effect = ship_side_effect
        ships_data = json.load(
            open(
                os.path.join(DATA_DIR, 'fleet_list_ships.json'),
                encoding='utf8'
            )
        )
        ships_side_effect = ShipsSideEffect(ships_data)
        ships_route = respx_mock.get('/my/ships')
        ships_route.side_effect = ships_side_effect

        survey_data = json.load(
            open(os.path.join(DATA_DIR, 'survey.json'), encoding='utf8')
        )
        expires = str(datetime.now(timezone.utc) + timedelta(hours=1))
        for survey in survey_data['data']['surveys']:
            survey['expiration'] = expires
        survey_side_effect = SurveySideEffect(survey_data)
        survey_route = respx_mock.post('/my/ships/TEST_SHIP_SYMBOL/survey')
        survey_route.side_effect = survey_side_effect

        # Good survey
        with ships_side_effect as tmp:
            tmp.data['data'][0]['symbol'] = 'TEST_SHIP_SYMBOL'
            tmp.data['data'][0]['nav']['systemSymbol'] = 'TEST-SYSTEM'
            tmp.data['data'][0]['nav']['waypointSymbol'] = 'TEST-SYSTEM-WAYPOINT'  # noqa: E501
            tmp.data['data'][0]['nav']['route']['destination']['type'] = 'ASTEROID'  # noqa: E501
            tmp.data['data'][0]['nav']['status'] = 'IN_ORBIT'
            tmp.data['data'][0]['mounts'][0]['symbol'] = 'MOUNT_SURVEYOR_II'

            ship = self.agent.fleet('TEST_SHIP_SYMBOL')
            surveys = ship.waypoints.surveys()
            assert inspect.isgenerator(surveys)
            surveys = list(surveys)
            assert len(surveys) == 1

        # Failed survey
        with ships_side_effect as ship_tmp, survey_side_effect as survey_tmp:
            ship_tmp.data['data'][0]['symbol'] = 'TEST_SHIP_SYMBOL'
            ship_tmp.data['data'][0]['nav']['systemSymbol'] = 'TEST-SYSTEM'
            ship_tmp.data['data'][0]['nav']['waypointSymbol'] = 'TEST-SYSTEM-WAYPOINT'  # noqa: E501
            ship_tmp.data['data'][0]['nav']['route']['destination']['type'] = 'ASTEROID'  # noqa: E501
            ship_tmp.data['data'][0]['nav']['status'] = 'IN_ORBIT'
            ship_tmp.data['data'][0]['mounts'][0]['symbol'] = 'MOUNT_SURVEYOR_II'  # noqa: E501
            survey_tmp.no_mounts = True

            ship = self.agent.fleet('TEST_SHIP_SYMBOL')
            surveys = list(ship.waypoints.surveys())

        # Only one waypoint surveyed
        with ships_side_effect as tmp:
            tmp.data['data'][0]['symbol'] = 'TEST_SHIP_SYMBOL'
            tmp.data['data'][0]['nav']['systemSymbol'] = 'TEST-SYSTEM'
            tmp.data['data'][0]['nav']['waypointSymbol'] = 'TEST-SYSTEM-WAYPOINT'  # noqa: E501
            tmp.data['data'][0]['nav']['route']['destination']['type'] = 'ASTEROID'  # noqa: E501
            tmp.data['data'][0]['nav']['status'] = 'IN_ORBIT'
            tmp.data['data'][0]['mounts'][0]['symbol'] = 'MOUNT_SURVEYOR_II'
            tmp.data['data'][1]['symbol'] = 'TEST_SHIP_SYMBOL'
            tmp.data['data'][1]['nav']['systemSymbol'] = 'TEST-SYSTEM'
            tmp.data['data'][1]['nav']['waypointSymbol'] = 'TEST-SYSTEM-WAYPOINT'  # noqa: E501
            tmp.data['data'][1]['nav']['route']['destination']['type'] = 'ASTEROID'  # noqa: E501
            tmp.data['data'][1]['nav']['status'] = 'IN_ORBIT'
            tmp.data['data'][1]['mounts'][0]['symbol'] = 'MOUNT_SURVEYOR_II'

            ship = self.agent.fleet('TEST_SHIP_SYMBOL')
            surveys = ship.waypoints.surveys()
            assert inspect.isgenerator(surveys)
            surveys = list(surveys)
            assert len(surveys) == 1

    @pytest.mark.respx(base_url='https://api.spacetraders.io/v2')
    def test_iter(self, respx_mock):
        ship_data = json.load(
            open(os.path.join(DATA_DIR, 'ship_info.json'), encoding='utf8')
        )
        ship_data['data']['nav']['systemSymbol'] = 'TEST-SYSTEM'
        ship_data['data']['nav']['waypointSymbol'] = 'TEST-SYSTEM-WAYPOINT'
        respx_mock.get('/my/ships/TEST_SHIP_SYMBOL').mock(
            return_value=httpx.Response(200, json=ship_data)
        )
        waypoints_data = json.load(
            open(os.path.join(DATA_DIR, 'waypoints.json'), encoding='utf8')
        )
        respx_mock.get(
            '/systems/TEST-SYSTEM/waypoints', params={'page': 1}
        ).mock(
            return_value=httpx.Response(200, json=waypoints_data)
        )
        respx_mock.get(
            '/systems/TEST-SYSTEM/waypoints', params={'page': 2}
        ).mock(
            return_value=httpx.Response(200, json={'data': []})
        )
        ship = self.agent.fleet('TEST_SHIP_SYMBOL')
        waypoints = list(ship.waypoints)
        assert len(waypoints) == 3

    @pytest.mark.respx(base_url='https://api.spacetraders.io/v2')
    def test_call(self, respx_mock):
        ship_data = json.load(
            open(os.path.join(DATA_DIR, 'ship_info.json'), encoding='utf8')
        )
        ship_data['data']['nav']['systemSymbol'] = 'TEST-SYSTEM'
        ship_data['data']['nav']['waypointSymbol'] = 'TEST-SYSTEM-WAYPOINT'
        respx_mock.get('/my/ships/TEST_SHIP_SYMBOL').mock(
            return_value=httpx.Response(200, json=ship_data)
        )
        waypoints_data = json.load(
            open(os.path.join(DATA_DIR, 'waypoints.json'), encoding='utf8')
        )
        respx_mock.get(
            '/systems/TEST-SYSTEM/waypoints', params={'page': 1}
        ).mock(
            return_value=httpx.Response(200, json=waypoints_data)
        )
        respx_mock.get(
            '/systems/TEST-SYSTEM/waypoints', params={'page': 2}
        ).mock(
            return_value=httpx.Response(200, json={'data': []})
        )
        ship = self.agent.fleet('TEST_SHIP_SYMBOL')
        waypoints = ship.waypoints()
        assert snisp.utils.ilen(waypoints) == 3
        waypoints = ship.waypoints(traits='UNCHARTED')
        assert snisp.utils.ilen(waypoints) == 3
        waypoints = ship.waypoints(types='PLANET')
        assert snisp.utils.ilen(waypoints) == 3
        waypoints = ship.waypoints(traits='UNCHARTED')
        assert snisp.utils.ilen(waypoints) == 3
        waypoints = ship.waypoints(filters={'isUnderConstruction': False})
        assert snisp.utils.ilen(waypoints) == 0
        waypoints = ship.waypoints(filters={'isUnderConstruction': True})
        assert snisp.utils.ilen(waypoints) == 3
        waypoints = ship.waypoints(system_symbol='TEST-SYSTEM')
        assert snisp.utils.ilen(waypoints) == 3

        with pytest.raises(snisp.exceptions.SpaceAttributeError):
            waypoints = next(ship.waypoints(traits='INVALID'))

        with pytest.raises(snisp.exceptions.SpaceAttributeError):
            waypoints = next(ship.waypoints(types='INVALID'))

    @pytest.mark.respx(base_url='https://api.spacetraders.io/v2')
    def test_get(self, respx_mock):
        ship_data = json.load(
            open(os.path.join(DATA_DIR, 'ship_info.json'), encoding='utf8')
        )
        ship_data['data']['nav']['systemSymbol'] = 'TEST-SYSTEM'
        ship_data['data']['nav']['waypointSymbol'] = 'TEST-SYSTEM-WAYPOINT'
        respx_mock.get('/my/ships/TEST_SHIP_SYMBOL').mock(
            return_value=httpx.Response(200, json=ship_data)
        )
        waypoint_data = json.load(
            open(os.path.join(DATA_DIR, 'waypoint.json'), encoding='utf8')
        )
        waypoint_side_effect = GenericSideEffect(waypoint_data)
        waypoint_route = respx_mock.get(
            '/systems/TEST-SYSTEM/waypoints/TEST-SYSTEM-WAYPOINT'
        )
        waypoint_route.side_effect = waypoint_side_effect

        ship = self.agent.fleet('TEST_SHIP_SYMBOL')
        waypoint = ship.waypoints.get()

        assert waypoint.to_dict() == waypoint_data['data']
        assert waypoint == ship.waypoints.get(waypoint=waypoint)
        assert waypoint == ship.waypoints.get(system_symbol='TEST-SYSTEM')
        assert waypoint == ship.waypoints.get(
            waypoint_symbol='TEST-SYSTEM-WAYPOINT'
        )
        assert waypoint == ship.waypoints.get(
            system_symbol='TEST-SYSTEM', waypoint_symbol='TEST-SYSTEM-WAYPOINT'
        )
        assert snisp.waypoints.is_uncharted(waypoint)

        with waypoint_side_effect as tmp:
            for trait in tmp.data['data']['traits']:
                if trait['symbol'] == 'UNCHARTED':
                    trait['symbol'] = 'CHARTED'
            waypoint = ship.waypoints.get()
            assert not snisp.waypoints.is_uncharted(waypoint)


class TestWaypoint:

    agent = snisp.agent.Agent(symbol='testing', faction='testing')

    @pytest.mark.respx(base_url='https://api.spacetraders.io/v2')
    def test_data(self, respx_mock):
        waypoint_data = json.load(
            open(os.path.join(DATA_DIR, 'waypoint.json'), encoding='utf8')
        )
        waypoint_data['data']['waypointSymbol'] = waypoint_data['data']['symbol']  # noqa: E501
        respx_mock.get(
            '/systems/TEST-SYSTEM/waypoints/TEST-SYSTEM-WAYPOINT'
        ).mock(
            return_value=httpx.Response(200, json=waypoint_data)
        )

        waypoint = snisp.waypoints.Waypoint(self.agent, waypoint_data['data'])

        assert waypoint.data.to_dict() == waypoint.to_dict()
        assert waypoint.data.to_dict() == waypoint_data['data']

    @pytest.mark.respx(base_url='https://api.spacetraders.io/v2')
    def test_location(self, respx_mock):
        waypoint_data = json.load(
            open(os.path.join(DATA_DIR, 'waypoint.json'), encoding='utf8')
        )
        waypoint_data['data']['waypointSymbol'] = waypoint_data['data']['symbol']  # noqa: E501

        waypoint = snisp.waypoints.Waypoint(self.agent, waypoint_data['data'])

        assert waypoint.location.sector == waypoint_data['data']['systemSymbol'].split('-')[0]  # noqa: E501
        assert waypoint.location.system == waypoint_data['data']['systemSymbol']  # noqa: E501
        assert waypoint.location.waypoint == waypoint_data['data']['symbol']

    @pytest.mark.respx(base_url='https://api.spacetraders.io/v2')
    def test_supply_construction_data(self, respx_mock):
        ship_data = json.load(
            open(os.path.join(DATA_DIR, 'ship_info.json'), encoding='utf8')
        )
        ship_data['data']['cargo']['capacity'] = 40
        dock_data = json.load(
            open(os.path.join(DATA_DIR, 'dock.json'), encoding='utf8')
        )
        respx_mock.post('/my/ships/TEST_SHIP_SYMBOL/dock').mock(
            return_value=httpx.Response(200, json=dock_data)
        )

        supply_data = json.load(
            open(
                os.path.join(DATA_DIR, 'supply_construction_site.json'),
                encoding='utf8'
            )
        )
        construction_data = json.load(
            open(
                os.path.join(DATA_DIR, 'construction_site.json'),
                encoding='utf8'
            )
        )
        construction_data['data']['systemSymbol'] = 'TEST-SYSTEM'
        construction_data['data']['isComplete'] = False
        construction_data['data']['systemSymbol'] = 'TEST-SYSTEM'
        construction_data['data']['materials'][0]['required'] = 40
        construction_data['data']['materials'][0]['fulfilled'] = 10

        supply_data['data']['systemSymbol'] = 'TEST-SYSTEM'
        supply_data['data']['construction']['materials'][0]['required'] = 40
        supply_data['data']['construction']['materials'][0]['fulfilled'] = 10
        supply_data['data']['cargo']['capacity'] = 40

        construction_side_effect = SupplyConstructionSiteSideEffect(
            data=supply_data,
            site_data=construction_data,
            ship_data=ship_data,
        )
        supply_route = respx_mock.post(
            '/systems/TEST-SYSTEM/waypoints/'
            'TEST-SYSTEM-CONSTRUCTION/construction/supply'
        )
        supply_route.side_effect = construction_side_effect

        ship_route = respx_mock.get('/my/ships/TEST_SHIP_SYMBOL')
        ship_route.side_effect = construction_side_effect.ship

        construction_site = snisp.waypoints.ConstructionSite(
            self.agent, construction_data['data']
        )

        # Supply until completed
        with construction_side_effect as tmp:
            tmp.ship_data['data']['cargo']['capacity'] = 40
            tmp.ship_data['data']['cargo']['units'] = 40
            tmp.ship_data['data']['cargo']['inventory'][0] = {
                'symbol': 'PRECIOUS_STONES', 'units': 40
            }
            ship = self.agent.fleet('TEST_SHIP_SYMBOL')
            while not construction_site.is_complete:
                prev_supply_count = construction_site.materials[0].fulfilled
                prev_cargo_count = ship.cargo.units
                construction_site.supply(
                    ship=ship, trade_symbol='PRECIOUS_STONES', units=5
                )
                assert ship.cargo.units == prev_cargo_count - 5
                assert construction_site.materials[0].fulfilled == prev_supply_count + 5  # noqa: E501
                prev_supply_count = construction_site.materials[0].fulfilled
                prev_cargo_count = ship.cargo.units
            assert construction_site.is_complete
            assert construction_site.materials[0].fulfilled == construction_site.materials[0].required  # noqa: E501
            assert ship.cargo.units == 10

        # 4800: ConstructionMaterialNotRequired
        with construction_side_effect as tmp:
            ship = self.agent.fleet('TEST_SHIP_SYMBOL')
            with pytest.raises(
                snisp.exceptions.ConstructionMaterialNotRequired
            ):
                construction_site.supply(
                    ship=ship, trade_symbol='INVALID', units=5
                )

        # 4801: ConstructionMaterialFulfilled
        with construction_side_effect as tmp:
            ship = self.agent.fleet('TEST_SHIP_SYMBOL')
            with pytest.raises(
                snisp.exceptions.ConstructionMaterialFulfilled
            ):
                construction_site.supply(
                    ship=ship, trade_symbol='PRECIOUS_STONES', units=50
                )

        # 4802: ShipConstructionInvalidLocationError
        with construction_side_effect as tmp:
            tmp.invalid = True
            ship = self.agent.fleet('TEST_SHIP_SYMBOL')
            with pytest.raises(
                snisp.exceptions.ShipConstructionInvalidLocationError
            ):
                construction_site.supply(
                    ship=ship, trade_symbol='PRECIOUS_STONES', units=50
                )

    @pytest.mark.respx(base_url='https://api.spacetraders.io/v2')
    def test_construction_data(self, respx_mock):
        construction_data = json.load(
            open(
                os.path.join(DATA_DIR, 'construction_site.json'),
                encoding='utf8'
            )
        )
        construction_data['data']['systemSymbol'] = 'TEST-SYSTEM'
        respx_mock.get(
            '/systems/TEST-SYSTEM/waypoints/'
            'TEST-SYSTEM-CONSTRUCTION/construction'
        ).mock(
            return_value=httpx.Response(200, json=construction_data)
        )

        construction_site = snisp.waypoints.ConstructionSite(
            self.agent, construction_data['data']
        )

        assert construction_site.to_dict() == construction_data['data']
        assert construction_site.refresh().to_dict() == construction_site.to_dict()  # noqa: E501
        assert construction_site.refresh().refresh().to_dict() == construction_site.to_dict()  # noqa: E501
        assert construction_site.refresh().to_dict() == construction_data['data']  # noqa: E501

    @pytest.mark.respx(base_url='https://api.spacetraders.io/v2')
    def test_refresh(self, respx_mock):
        ship_data = json.load(
            open(os.path.join(DATA_DIR, 'ship_info.json'), encoding='utf8')
        )
        ship_data['data']['nav']['systemSymbol'] = 'TEST-SYSTEM'
        ship_data['data']['nav']['waypointSymbol'] = 'TEST-SYSTEM-WAYPOINT'
        respx_mock.get('/my/ships/TEST_SHIP_SYMBOL').mock(
            return_value=httpx.Response(200, json=ship_data)
        )
        waypoint_data = json.load(
            open(os.path.join(DATA_DIR, 'waypoint.json'), encoding='utf8')
        )
        waypoint_side_effect = GenericSideEffect(waypoint_data)
        waypoint_route = respx_mock.get(
            '/systems/TEST-SYSTEM/waypoints/TEST-SYSTEM-WAYPOINT'
        )
        waypoint_route.side_effect = waypoint_side_effect

        ship = self.agent.fleet('TEST_SHIP_SYMBOL')
        waypoint = ship.waypoints.get()

        assert waypoint.to_dict() == ship.waypoints.refresh(waypoint).to_dict()

    @pytest.mark.respx(base_url='https://api.spacetraders.io/v2')
    def test_search(self, respx_mock):
        ship_data = json.load(
            open(os.path.join(DATA_DIR, 'ship_info.json'), encoding='utf8')
        )
        ship_data['data']['nav']['systemSymbol'] = 'TEST-SYSTEM'
        ship_data['data']['nav']['waypointSymbol'] = 'TEST-SYSTEM-WAYPOINT'
        respx_mock.get('/my/ships/TEST_SHIP_SYMBOL').mock(
            return_value=httpx.Response(200, json=ship_data)
        )
        waypoints_data = json.load(
            open(os.path.join(DATA_DIR, 'waypoints.json'), encoding='utf8')
        )
        respx_mock.get(
            '/systems/TEST-SYSTEM/waypoints', params={'page': 1}
        ).mock(
            return_value=httpx.Response(200, json=waypoints_data)
        )
        respx_mock.get(
            '/systems/TEST-SYSTEM/waypoints', params={'page': 2}
        ).mock(
            return_value=httpx.Response(200, json={'data': []})
        )

        ship = self.agent.fleet('TEST_SHIP_SYMBOL')
        waypoints = ship.waypoints.search()
        assert snisp.utils.ilen(waypoints) == 3
        waypoints = ship.waypoints.search(traits='UNCHARTED')
        assert snisp.utils.ilen(waypoints) == 3
        waypoints = ship.waypoints.search(types='PLANET')
        assert snisp.utils.ilen(waypoints) == 3
        waypoints = ship.waypoints.search(traits='UNCHARTED')
        assert snisp.utils.ilen(waypoints) == 3
        waypoints = ship.waypoints.search(
            filters={'isUnderConstruction': False}
        )
        assert snisp.utils.ilen(waypoints) == 0
        waypoints = ship.waypoints.search(
            filters={'isUnderConstruction': True}
        )
        assert snisp.utils.ilen(waypoints) == 3
        waypoints = ship.waypoints.search(system_symbol='TEST-SYSTEM')
        assert snisp.utils.ilen(waypoints) == 3

        with pytest.raises(snisp.exceptions.SpaceAttributeError):
            waypoints = next(ship.waypoints.search(traits='INVALID'))

        with pytest.raises(snisp.exceptions.SpaceAttributeError):
            waypoints = next(ship.waypoints.search(types='INVALID'))

    @pytest.mark.respx(base_url='https://api.spacetraders.io/v2')
    def test_by_types(self, respx_mock):
        ship_data = json.load(
            open(os.path.join(DATA_DIR, 'ship_info.json'), encoding='utf8')
        )
        ship_data['data']['nav']['systemSymbol'] = 'TEST-SYSTEM'
        ship_data['data']['nav']['waypointSymbol'] = 'TEST-SYSTEM-WAYPOINT'
        respx_mock.get('/my/ships/TEST_SHIP_SYMBOL').mock(
            return_value=httpx.Response(200, json=ship_data)
        )
        waypoints_data = json.load(
            open(os.path.join(DATA_DIR, 'waypoints.json'), encoding='utf8')
        )
        waypoints_side_effect = WaypointsSideEffect(waypoints_data)
        waypoints_route = respx_mock.get('/systems/TEST-SYSTEM/waypoints')
        waypoints_route.side_effect = waypoints_side_effect

        construction_data = json.load(
            open(
                os.path.join(DATA_DIR, 'construction_site.json'),
                encoding='utf8'
            )
        )
        for waypoint in waypoints_data['data']:
            respx_mock.get(
                '/systems/TEST-SYSTEM/waypoints/'
                f'{waypoint["symbol"]}/construction'
            ).mock(
                return_value=httpx.Response(200, json=construction_data)
            )

        ship = self.agent.fleet('TEST_SHIP_SYMBOL')

        with waypoints_side_effect as tmp:
            assert snisp.utils.ilen(
                ship.waypoints.construction_sites()
            ) == 3

        with waypoints_side_effect as tmp:
            tmp.data['data'][-1]['traits'][0]['symbol'] = 'SHIPYARD'
            assert snisp.utils.ilen(ship.waypoints.shipyards()) == 1

        with waypoints_side_effect as tmp:
            assert snisp.utils.ilen(ship.waypoints.shipyards()) == 0

        with waypoints_side_effect as tmp:
            assert snisp.utils.ilen(ship.waypoints.planets()) == 3

        with waypoints_side_effect as tmp:
            for wp in tmp.data['data']:
                wp['type'] = 'INVALID'
            assert snisp.utils.ilen(ship.waypoints.planets()) == 0

        with waypoints_side_effect as tmp:
            tmp.data['data'][-1]['type'] = 'GAS_GIANT'
            assert snisp.utils.ilen(ship.waypoints.gas_giants()) == 1

        with waypoints_side_effect as tmp:
            for wp in tmp.data['data']:
                wp['type'] = 'INVALID'
            assert snisp.utils.ilen(ship.waypoints.gas_giants()) == 0

        with waypoints_side_effect as tmp:
            tmp.data['data'][-1]['type'] = 'MOON'
            assert snisp.utils.ilen(ship.waypoints.moons()) == 1

        with waypoints_side_effect as tmp:
            for wp in tmp.data['data']:
                wp['type'] = 'INVALID'
            assert snisp.utils.ilen(ship.waypoints.moons()) == 0

        with waypoints_side_effect as tmp:
            tmp.data['data'][-1]['type'] = 'ORBITAL_STATION'
            assert snisp.utils.ilen(ship.waypoints.orbital_stations()) == 1

        with waypoints_side_effect as tmp:
            for wp in tmp.data['data']:
                wp['type'] = 'INVALID'
            assert snisp.utils.ilen(ship.waypoints.orbital_stations()) == 0

        with waypoints_side_effect as tmp:
            tmp.data['data'][-1]['type'] = 'JUMP_GATE'
            respx_mock.get(
                '/systems/TEST-SYSTEM/waypoints/'
                'TEST-SYSTEM-FARTHESTWAYPOINT/jump-gate'
            ).mock(
                return_value=httpx.Response(
                    200, json={'data': tmp.data['data'][-1]}
                )
            )
            assert snisp.utils.ilen(ship.waypoints.jump_gates()) == 1
            jump_gate = next(ship.waypoints.jump_gates())
            assert jump_gate.data.to_dict() == jump_gate.to_dict()

        with waypoints_side_effect as tmp:
            for wp in tmp.data['data']:
                wp['type'] = 'INVALID'
            assert snisp.utils.ilen(ship.waypoints.jump_gates()) == 0

        with waypoints_side_effect as tmp:
            tmp.data['data'][-1]['type'] = 'ASTEROID_FIELD'
            assert snisp.utils.ilen(ship.waypoints.asteroid_fields()) == 1

        with waypoints_side_effect as tmp:
            for wp in tmp.data['data']:
                wp['type'] = 'INVALID'
            assert snisp.utils.ilen(ship.waypoints.asteroid_fields()) == 0

        with waypoints_side_effect as tmp:
            tmp.data['data'][-1]['type'] = 'ASTEROID'
            assert snisp.utils.ilen(ship.waypoints.asteroids()) == 1

        with waypoints_side_effect as tmp:
            for wp in tmp.data['data']:
                wp['type'] = 'INVALID'
            assert snisp.utils.ilen(ship.waypoints.asteroids()) == 0

        with waypoints_side_effect as tmp:
            tmp.data['data'][-1]['type'] = 'ENGINEERED_ASTEROID'
            assert snisp.utils.ilen(ship.waypoints.engineered_asteroids()) == 1

        with waypoints_side_effect as tmp:
            for wp in tmp.data['data']:
                wp['type'] = 'INVALID'
            assert snisp.utils.ilen(ship.waypoints.engineered_asteroids()) == 0

        with waypoints_side_effect as tmp:
            tmp.data['data'][-1]['type'] = 'ASTEROID_BASE'
            assert snisp.utils.ilen(ship.waypoints.asteroid_bases()) == 1

        with waypoints_side_effect as tmp:
            for wp in tmp.data['data']:
                wp['type'] = 'INVALID'
            assert snisp.utils.ilen(ship.waypoints.asteroid_bases()) == 0

        with waypoints_side_effect as tmp:
            tmp.data['data'][-1]['type'] = 'NEBULA'
            assert snisp.utils.ilen(ship.waypoints.nebulas()) == 1

        with waypoints_side_effect as tmp:
            for wp in tmp.data['data']:
                wp['type'] = 'INVALID'
            assert snisp.utils.ilen(ship.waypoints.nebulas()) == 0

        with waypoints_side_effect as tmp:
            tmp.data['data'][-1]['type'] = 'DEBRIS_FIELD'
            assert snisp.utils.ilen(ship.waypoints.debris_fields()) == 1

        with waypoints_side_effect as tmp:
            for wp in tmp.data['data']:
                wp['type'] = 'INVALID'
            assert snisp.utils.ilen(ship.waypoints.debris_fields()) == 0

        with waypoints_side_effect as tmp:
            tmp.data['data'][-1]['type'] = 'GRAVITY_WELL'
            assert snisp.utils.ilen(ship.waypoints.gravity_wells()) == 1

        with waypoints_side_effect as tmp:
            for wp in tmp.data['data']:
                wp['type'] = 'INVALID'
            assert snisp.utils.ilen(ship.waypoints.gravity_wells()) == 0

        with waypoints_side_effect as tmp:
            tmp.data['data'][-1]['type'] = 'ARTIFICIAL_GRAVITY_WELL'
            assert snisp.utils.ilen(
                ship.waypoints.artificial_gravity_wells()
            ) == 1

        with waypoints_side_effect as tmp:
            tmp.data['data'][-1]['type'] = 'FUEL_STATION'
            assert snisp.utils.ilen(
                (i for i in ship.waypoints if i.type == 'FUEL_STATION')
            ) == 1

        with waypoints_side_effect as tmp:
            for wp in tmp.data['data']:
                wp['type'] = 'INVALID'
            assert snisp.utils.ilen(
                ship.waypoints.artificial_gravity_wells()
            ) == 0

    def test_class_factory(self, respx_mock):
        waypoint_data = json.load(
            open(os.path.join(DATA_DIR, 'waypoint.json'), encoding='utf8')
        )
        cls = snisp.waypoints.class_factory('Waypoint')
        waypoint = cls(self.agent, waypoint_data['data'])
        assert waypoint.to_dict() == waypoint_data['data']
        cls = snisp.waypoints.class_factory('FakeWaypointType')
        waypoint = cls(self.agent, waypoint_data['data'])
        assert waypoint.to_dict() == waypoint_data['data']

    @pytest.mark.respx(base_url='https://api.spacetraders.io/v2')
    def test_best_asteroid(self, respx_mock):
        ships_data = json.load(
            open(
                os.path.join(DATA_DIR, 'fleet_list_ships.json'),
                encoding='utf8'
            )
        )
        ships_data['data'][0]['symbol'] = 'TEST_SHIP_SYMBOL'
        ships_data['data'][0]['nav']['systemSymbol'] = 'TEST-SYSTEM'
        ships_data['data'][0]['nav']['waypointSymbol'] = 'TEST-SYSTEM-WAYPOINT'
        ships_data['data'][0]['nav']['status'] = 'IN_ORBIT'
        ships_data['data'][0]['nav']['route']['destination']['type'] = 'ASTEROID'  # noqa: E501
        ships_data['data'][0]['mounts'][0]['symbol'] = 'MOUNT_SURVEYOR_II'
        ship_data = json.load(
            open(os.path.join(DATA_DIR, 'ship_info.json'), encoding='utf8')
        )
        ship_data['data']['nav']['systemSymbol'] = 'TEST-SYSTEM'
        ship_data['data']['nav']['waypointSymbol'] = 'TEST-SYSTEM-WAYPOINT'
        ship_data['data']['nav']['status'] = 'IN_ORBIT'
        ship_data['data']['nav']['route']['destination']['type'] = 'ASTEROID'
        respx_mock.get(
                '/my/ships', params={'page': 1, 'limit': 20}
        ).mock(
            return_value=httpx.Response(200, json=ships_data)
        )
        respx_mock.get(
                '/my/ships', params={'page': 2, 'limit': 20}
        ).mock(
            return_value=httpx.Response(200, json={'data': []})
        )
        respx_mock.get('/my/ships/TEST_SHIP_SYMBOL').mock(
            return_value=httpx.Response(200, json=ship_data)
        )

        contracts_data = json.load(
            open(os.path.join(DATA_DIR, 'contracts.json'), encoding='utf8')
        )
        survey_data = json.load(
            open(os.path.join(DATA_DIR, 'survey.json'), encoding='utf8')
        )
        waypoint_data = json.load(
            open(os.path.join(DATA_DIR, 'waypoint.json'), encoding='utf8')
        )
        waypoint_data['data']['type'] = 'ASTEROID'

        ship = self.agent.fleet('TEST_SHIP_SYMBOL')

        best_asteroid_side_effect = BestAsteroidSideEffect(
            contracts=contracts_data,
            surveys=survey_data,
            waypoint=waypoint_data,
        )

        waypoint_route = respx_mock.route(
            M(path__regex=r'/systems/TEST-SYSTEM/waypoints/.*')
        )
        waypoint_route.side_effect = best_asteroid_side_effect.waypoint

        contracts_route = respx_mock.get('/my/contracts')
        contracts_route.side_effect = best_asteroid_side_effect.contracts

        survey_route = respx_mock.post('/my/ships/TEST_SHIP_SYMBOL/survey')
        survey_route.side_effect = best_asteroid_side_effect.survey

        asteroid = snisp.waypoints.best_asteroid(ship)
        assert asteroid.type in snisp.utils.SURVEYABLE_WAYPOINTS

        asteroid = snisp.waypoints.best_asteroid(ship, 'IRON_ORE')
        assert asteroid.type in snisp.utils.SURVEYABLE_WAYPOINTS

        assert not snisp.waypoints.best_asteroid(ship, 'INVALID')

        with best_asteroid_side_effect as tmp:
            for survey in tmp.surveys_data['data']['surveys']:
                survey['size'] = 'MODERATE'
            asteroid = snisp.waypoints.best_asteroid(ship)
            assert asteroid.type in snisp.utils.SURVEYABLE_WAYPOINTS

        with best_asteroid_side_effect as tmp:
            for survey in tmp.surveys_data['data']['surveys']:
                survey['size'] = 'SMALL'
            asteroid = snisp.waypoints.best_asteroid(ship)
            assert asteroid.type in snisp.utils.SURVEYABLE_WAYPOINTS

        with best_asteroid_side_effect as tmp:
            tmp.contracts_data['data'][-1]['terms']['deliver'][0]['tradeSymbol'] = 'IRON_ORE'  # noqa: E501
            tmp.contracts_data['data'][-1]['terms']['deliver'][0]['unitsFulfilled'] = 0  # noqa: E501
            tmp.contracts_data['data'][-1]['terms']['deliver'][0]['unitsRequired'] = 100  # noqa: E501
            contract = self.agent.contracts.current
            assert snisp.waypoints.best_asteroid(ship, contract=contract)


class ChartSideEffect:

    def __init__(self, data):
        self.orig_data = data
        self.data = copy.deepcopy(self.orig_data)
        self.already_charted = False
        self.fail = False

    def __call__(self, request, route):
        if self.already_charted:
            return httpx.Response(
                400, json={'error': {'data': {'code': 4230}}}
            )
        if self.fail:
            return httpx.Response(400, json={})
        return httpx.Response(200, json=self.data)

    def __enter__(self):
        self.reset()
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.reset()

    def reset(self):
        self.data = copy.deepcopy(self.orig_data)
        self.already_charted = False
        self.fail = False


class SurveySideEffect:

    def __init__(self, data):
        self.orig_data = data
        self.data = copy.deepcopy(self.orig_data)
        self.no_mounts = False

    def __call__(self, request, route):
        if self.no_mounts:
            return httpx.Response(
                400, json={'error': {'data': {'code': 4240}}}
            )
        return httpx.Response(200, json=self.data)

    def __enter__(self):
        self.reset()
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.reset()

    def reset(self):
        self.data = copy.deepcopy(self.orig_data)
        self.no_mounts = False


class ShipsSideEffect:

    def __init__(self, data):
        self.orig_data = data
        self.data = copy.deepcopy(self.orig_data)

    def __call__(self, request, route):
        if int(request.url.params.get('page', 1)) > 1:
            return httpx.Response(200, json={'data': []})
        return httpx.Response(200, json=self.data)

    def __enter__(self):
        self.reset()
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.reset()

    def reset(self):
        self.data = copy.deepcopy(self.orig_data)


class WaypointsSideEffect:

    def __init__(self, data):
        self.orig_data = data
        self.data = copy.deepcopy(self.orig_data)

    def __call__(self, request, route):
        if int(request.url.params.get('page', 1)) > 1:
            return httpx.Response(200, json={'data': []})
        if traits := request.url.params.get('traits'):
            output = []
            for wp in self.data['data']:
                if any(i['symbol'] == traits for i in wp['traits']):
                    output.append(wp)
            self.data['data'] = output
        if types := request.url.params.get('type'):
            output = []
            for wp in self.data['data']:
                if types == wp['type']:
                    output.append(wp)
            self.data['data'] = output
        return httpx.Response(200, json=self.data)

    def __enter__(self):
        self.reset()
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.reset()

    def reset(self):
        self.data = copy.deepcopy(self.orig_data)


class SupplyConstructionSiteSideEffect:

    def __init__(self, *, data, site_data, ship_data):
        self.orig_data = data
        self.data = copy.deepcopy(self.orig_data)
        self.orig_site_data = site_data
        self.site_data = copy.copy(self.orig_site_data)
        self.orig_ship_data = ship_data
        self.ship_data = copy.deepcopy(self.orig_ship_data)
        self.invalid = False

    def __call__(self, request, route):
        if self.invalid:
            return httpx.Response(
                400, json={'error': {'data': {'code': 4802}}}
            )
        payload = json.loads(request.content.decode('utf8'))
        units = int(payload['units'])
        material = next(
            (
                m for m in self.data['data']['construction']['materials']
                if payload['tradeSymbol'] == m['tradeSymbol']
            ), None
        )
        if material is None:
            return httpx.Response(
                400, json={'error': {'data': {'code': 4800}}}
            )
        if units > material['required']:
            return httpx.Response(
                400, json={'error': {'data': {'code': 4801}}}
            )
        material['fulfilled'] += units
        self.ship_data['data']['cargo']['units'] -= units
        out_inventory = []
        for item in self.ship_data['data']['cargo']['inventory']:
            if item['symbol'] == payload['tradeSymbol']:
                item['units'] -= units
                if item['units'] > 0:
                    out_inventory.append(item)
        if all(
            m['required'] == m['fulfilled']
            for m in self.data['data']['construction']['materials']
        ):
            self.data['data']['construction']['isComplete'] = True
            self.site_data['data']['isComplete'] = True
        self.ship_data['data']['cargo']['inventory'] = out_inventory
        self.data['data']['cargo'] = self.ship_data['data']['cargo']
        return httpx.Response(200, json=self.data)

    def ship(self, request, route):
        return httpx.Response(200, json=self.ship_data)

    def __enter__(self):
        self.reset()
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.reset()

    def reset(self):
        self.data = copy.deepcopy(self.orig_data)
        self.site_data = copy.copy(self.orig_site_data)
        self.ship_data = copy.deepcopy(self.orig_ship_data)
        self.invalid = False


class BestAsteroidSideEffect:

    def __init__(self, *, contracts, surveys, waypoint):
        self.orig_contracts = contracts
        self.contracts_data = copy.deepcopy(self.orig_contracts)
        self.orig_surveys = surveys
        self.surveys_data = copy.deepcopy(self.orig_surveys)
        self.orig_waypoint = waypoint
        self.waypoint_data = copy.deepcopy(self.orig_waypoint)

    def __enter__(self):
        self.reset()
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.reset()

    def contracts(self, request, route):
        if int(request.url.params.get('page', 1)) > 1:
            return httpx.Response(200, json={'data': []})
        return httpx.Response(200, json=self.contracts_data)

    def survey(self, request, route):
        return httpx.Response(200, json=self.surveys_data)

    def waypoint(self, request, route):
        return httpx.Response(200, json=self.waypoint_data)

    def reset(self):
        self.contracts_data = copy.deepcopy(self.orig_contracts)
        self.surveys_data = copy.deepcopy(self.orig_surveys)
        self.waypoint_data = copy.deepcopy(self.orig_waypoint)



import itertools
import logging

from collections import namedtuple

from snisp import cache, utils
from snisp.exceptions import ClientError
from snisp.decorators import retry
from snisp.waypoints import Waypoints
from snisp.systems import Location


logger = logging.getLogger(__name__)

MarketDataRecord = namedtuple(
    'MarketDataRecord',
    [
        'distance',
        'trade_symbol',
        'import_market',
        'export_market',
        'import_waypoint',
        'export_waypoint',
    ]
)


class Markets:

    """Markets in the Ship's current Location"""

    # TODO: using __call__ for the actual market data is kind of confusing
    #       when used by a ship, e.g., ship.markets()
    #       Need to think of a better scheme/name

    def __init__(self, ship, location):
        self.ship = ship
        self.agent = ship.agent
        self.location = location

    def __repr__(self):
        cls = self.__class__.__name__
        return f'{cls}({self.agent!r}, {self.location!r})'

    def __iter__(self):
        """
        Iterates over the Waypoint's in the Ship's current location
        yielding Markets

        Yields:
            Market
        """
        _waypoint = Waypoints(self.agent, self.location)
        for waypoint in _waypoint(traits='MARKETPLACE'):
            yield Market(self.agent, waypoint.to_dict())

    @retry()
    def __call__(
        self, *, waypoint=None, waypoint_symbol=None
    ):
        """
        Retrieves a Market. Can be useful when going from
        Waypoint --> Market.

        If no kwarg is passed, the current Market data will be returned, which
        is equivalent to MarketData

        Kwargs:
            waypoint: A Waypoint or Waypoint sublcass
            waypoint_symbol: The string of a Waypoint that is a Market

        Returns:
            MarketData
        """
        if waypoint is not None:
            system_symbol = waypoint.system_symbol
            waypoint_symbol = waypoint.symbol
        elif waypoint_symbol is not None:
            system_symbol = '-'.join(waypoint_symbol.split('-')[:2])
        else:
            waypoint_symbol = self.location.waypoint
            system_symbol = self.location.system
        try:
            response = self.agent.client.get(
                f'/systems/{system_symbol}/waypoints/{waypoint_symbol}/market'
            )
        except ClientError as e:
            if data := e.data:
                if data.get('code') == 404:
                    logger.warning(
                        f'Waypoint at {system_symbol}-->{waypoint_symbol} '
                        'does not exist'
                    )
                    return MarketData(self.agent, {})
            raise e
        data = response.json()['data']
        data['location'] = Location(
            self.agent, {'headquarters': waypoint_symbol}
        )
        if not data.get('tradeGoods'):
            data['tradeGoods'] = []
        return MarketData(self.agent, data)

    def cheapest(self, trade_symbol):
        """
        Searches Markets in the Ship's system to return the Market
        with the cheapest Purchase Price for the trade symbol

        A Ship or Probe must be at the Market to get the current
        Sell Price. Otherwise, the closest Market that offers the trade
        symbol will be returned

        Args:
            trade_symbol: The symbol of the good to search

        Returns:
            Market if a Market that sells trade_symbol is found; else None
        """
        trade_symbol = trade_symbol.upper().strip()
        best_market = None
        best_price = None
        for market in self:
            market_data = market.data
            goods = itertools.chain(
                market_data.imports, market_data.exports, market_data.exchange
            )
            for good in goods:
                if good.symbol == trade_symbol:
                    purchase_price = next(
                        (
                            i.purchase_price
                            for i in market.data.trade_goods
                            if i.symbol == good.symbol
                        ), None
                    )
                    if best_market is None:
                        best_market = market
                        best_price = purchase_price
                    elif purchase_price:
                        # A market w/o a ship located at it will report that
                        # it sells the good but won't have the sell price
                        if not best_price or purchase_price < best_price:
                            best_market = market
                            best_price = purchase_price
        return best_market

    def cheapest_import(self, trade_symbol):
        """
        Searches Markets in the Ship's system to return the Market
        importing the trade symbol with the cheapest Sell Price

        A Ship or Probe must be at the Market to get the current
        Sell Price. Otherwise, the closest Market that imports the trade
        symbol will be returned

        Args:
            trade_symbol: The symbol of the good to search

        Returns:
            Market if a Market that imports trade_symbol is found; else None
        """
        if imports := list(self.search(imports=trade_symbol)):
            try:
                imports = sorted(
                    imports, key=lambda x: import_sort_key(x[1], trade_symbol)
                )
                best = import_sort_key(imports[0][1], trade_symbol)
                filtered = filter(
                    lambda x: import_sort_key(x[1], trade_symbol) == best,
                    imports
                )
                return min(filtered, key=lambda x: self.ship.distance(x[0]))
            except TypeError:
                return min(imports, key=lambda x: self.ship.distance(x[0]))

    def most_expensive_import(self, trade_symbol):
        """
        Searches Markets in the Ship's system to return the Market
        importing the trade symbol with the highest Sell Price

        A Ship or Probe must be at the Market to get the current
        Sell Price. Otherwise, the closest Market that imports the trade
        symbol will be returned

        Args:
            trade_symbol: The symbol of the good to search

        Returns:
            Market if a Market that imports trade_symbol is found; else None
        """
        if imports := list(self.search(imports=trade_symbol)):
            try:
                imports = sorted(
                    imports,
                    key=lambda x: x and import_sort_key(x[1], trade_symbol),
                    reverse=True,
                )
                best = import_sort_key(imports[0][1], trade_symbol)
                filtered = filter(
                    lambda x: import_sort_key(x[1], trade_symbol) == best,
                    imports
                )
                return min(filtered, key=lambda x: self.ship.distance(x[0]))
            except TypeError:
                return min(imports, key=lambda x: self.ship.distance(x[0]))

    def cheapest_export(self, trade_symbol):
        """
        Searches Markets in the Ship's system to return the Market
        exporting the trade symbol with the cheapest Purchase Price

        A Ship or Probe must be at the Market to get the current
        Sell Price. Otherwise, the closest Market that exports the trade
        symbol will be returned

        Args:
            trade_symbol: The symbol of the good to search

        Returns:
            Market if a Market that expports trade_symbol is found; else None
        """
        if exports := list(self.search(exports=trade_symbol)):
            try:
                exports = sorted(
                    exports, key=lambda x: export_sort_key(x[1], trade_symbol)
                )
                best = export_sort_key(exports[0][1], trade_symbol)
                filtered = filter(
                    lambda x: export_sort_key(x[1], trade_symbol) == best,
                    exports
                )
                return min(filtered, key=lambda x: self.ship.distance(x[0]))
            except TypeError:
                return min(exports, key=lambda x: self.ship.distance(x[0]))

    def cheapest_exchange(self, trade_symbol):
        """
        Searches Markets in the Ship's system to return the Market
        exchanging the trade symbol with the cheapest Purchase Price

        A Ship or Probe must be at the Market to get the current
        Sell Price. Otherwise, the closest Market that exchanges the trade
        symbol will be returned

        Args:
            trade_symbol: The symbol of the good to search

        Returns:
            Market if a Market that exchanges trade_symbol is found; else None
        """
        if exchanges := list(self.search(exchanges=trade_symbol)):
            try:
                exchanges = sorted(
                    exchanges,
                    key=lambda x: exchange_sort_key(x[1], trade_symbol)
                )
                best = exchange_sort_key(exchanges[0][1], trade_symbol)
                filtered = filter(
                    lambda x: exchange_sort_key(x[1], trade_symbol) == best,
                    exchanges
                )
                return min(filtered, key=lambda x: self.ship.distance(x[0]))
            except TypeError:
                return min(exchanges, key=lambda x: self.ship.distance(x[0]))

    def cheapest_export_or_exchange(self, trade_symbol):
        """
        Searches Markets in the Ship's system to return the Market
        exporting or exchanging the trade symbol with the
        cheapest Purchase Price

        A Ship or Probe must be at the Market to get the current
        Sell Price. Otherwise, the closest Market that exports or exchanges
        the trade symbol will be returned

        Args:
            trade_symbol: The symbol of the good to search

        Returns:
            Market if a Market that exports or exchanges trade_symbol is
            found; else None
        """
        markets = []
        for market in self:
            market_data = self(waypoint=market)
            if trade_symbol in {i.symbol for i in market_data.exports}:
                markets.append((market, market_data))
            elif trade_symbol in {i.symbol for i in market_data.exchange}:
                markets.append((market, market_data))
        if markets:
            try:
                markets = sorted(
                    markets,
                    key=lambda x: export_or_exchange_sort_key(
                        x[1], trade_symbol
                    )
                )
                best = export_or_exchange_sort_key(markets[0][1], trade_symbol)
                filtered = filter(
                    lambda x: export_or_exchange_sort_key(
                        x[1], trade_symbol
                    ) == best,
                    markets
                )
                return min(filtered, key=lambda x: self.ship.distance(x[0]))
            except TypeError:
                return min(markets, key=lambda x: self.ship.distance(x[0]))

    def most_expensive_export(self, trade_symbol):
        """
        Searches Markets in the Ship's system to return the Market
        exporting the trade symbol with the most expensive Purchase Price

        A Ship or Probe must be at the Market to get the current
        Sell Price. Otherwise, the closest Market that exports the trade
        symbol will be returned

        Args:
            trade_symbol: The symbol of the good to search

        Returns:
            Market if a Market that imports trade_symbol is found; else None
        """
        if exports := list(self.search(exports=trade_symbol)):
            try:
                exports = sorted(
                    exports,
                    key=lambda x: export_sort_key(x[1], trade_symbol),
                    reverse=True,
                )
                best = export_sort_key(exports[0][1], trade_symbol)
                filtered = filter(
                    lambda x: export_sort_key(x[1], trade_symbol) == best,
                    exports
                )
                return min(filtered, key=lambda x: self.ship.distance(x[0]))
            except TypeError:
                return min(exports, key=lambda x: self.ship.distance(x[0]))

    def search(self, imports=None, exports=None, exchanges=None):
        """
        Searches for Markets in the current System. Multiple arguments
        will be ANDs

        Kwargs:
            imports: Find a Market that imports trade symbol
            exports: Find a Market that exports trade symbol
            exchanges: Find a Market that exchanges trade symbol

        Yields:
            A tuple of (Market, MarketData)
        """
        if imports is not None:
            imports = imports.upper().strip()
        if exports is not None:
            exports = exports.upper().strip()
        if exchanges is not None:
            exchanges = exchanges.upper().strip()
        for market in self:
            market_data = self(waypoint=market)
            market_imports = {i.symbol for i in market_data.imports}
            if not imports or imports in market_imports:
                market_exports = {i.symbol for i in market_data.exports}
                if not exports or exports in market_exports:
                    market_exchanges = {
                        i.symbol for i in market_data.exchange
                    }
                    if not exchanges or exchanges in market_exchanges:
                        yield market, market_data

    def shipyard_market_data(self, ship_type):
        """
        Returns a tuple of the Shipyard, ShipyardShip of the cheapest
        ShipyardShip of type ship type

        Args:
            ship_type: Ship type to search. See snisp.utils.SHIP_TYPES
                       for all supported ship types

        Returns:
            A tuple of (Shipyard, ShipyardShip) if ship_type is found; else,
            a tuple of (None, None)
        """
        # Reminder: Available ships will yield nothing if no probes
        #           are at the waypoint
        output = (
            (i, s) for i in self.ship.waypoints.shipyards()
            for s in i.available_ships(ship_type=ship_type)
        )
        try:
            return min(output, key=lambda x: x[1].purchase_price)
        except ValueError:
            return None, None

    def sells(self, trade_symbol):
        """
        Finds Markets in the current System that sell the trade symbol

        Args:
            trade_symbol: The symbol of the good to find

        Yields:
            Market
        """
        trade_symbol = trade_symbol.strip().upper()
        for market in self:
            market_data = market.data
            if any(i.symbol == trade_symbol for i in market_data.imports):
                yield market
            elif any(i.symbol == trade_symbol for i in market_data.exports):
                yield market
            elif any(i.symbol == trade_symbol for i in market_data.exchange):
                yield market

    def imports(self, trade_symbol):
        """
        Finds Markets in the current System that import the trade symbol

        Args:
            trade_symbol: The symbol of the good to find

        Yields:
            Market
        """
        trade_symbol = trade_symbol.strip().upper()
        for market in self:
            if imports := market.data.imports:
                for _import in imports:
                    if _import.symbol == trade_symbol:
                        yield market

    def exports(self, trade_symbol):
        """
        Finds Markets in the current System that export the trade symbol

        Args:
            trade_symbol: The symbol of the good to find

        Yields:
            Market
        """
        trade_symbol = trade_symbol.strip().upper()
        for market in self:
            if exports := market.data.exports:
                for export in exports:
                    if export.symbol == trade_symbol:
                        yield market

    def exchanges(self, trade_symbol):
        """
        Finds Markets in the current System that exchange the trade symbol

        Args:
            trade_symbol: The symbol of the good to find

        Yields:
            Market
        """
        trade_symbol = trade_symbol.strip().upper()
        for market in self:
            if exchanges := market.data.exchange:
                for exchange in exchanges:
                    if exchange.symbol == trade_symbol:
                        yield market

    def fuel_stations(self, *, system_symbol=None, traits=None):
        """
        Finds all Waypoints in the system that allow refueling

        NOTE: Since this is an expensive and time consuming call that requires
        a considerable amount of network calls, the results will be cached.

        Returns:
            Waypoints: List of Waypoints that accept refueling
        """
        with self.agent.lock:
            if not self.agent.client.testing:
                if fuel_stations := cache.get_fuel_stations(self.location):
                    return fuel_stations
            fuel_stations = []
            fs_waypoints = Waypoints(self.agent, self.location)
            fuel_station_symbols = set()
            for fuel_station in itertools.chain(
                fs_waypoints(
                    system_symbol=system_symbol,
                    types='FUEL_STATION',
                    traits=traits,
                ),
                (i[0] for i in self.search(exports='FUEL')),
                (i[0] for i in self.search(exchanges='FUEL')),
            ):
                if fuel_station.symbol not in fuel_station_symbols:
                    fuel_stations.append(fuel_station)
                    fuel_station_symbols.add(fuel_station.symbol)
            if not self.agent.client.testing:
                cache.insert_fuel_stations(self.location, fuel_stations)
        return fuel_stations


class Market(utils.AbstractJSONItem):

    def __init__(self, agent, market):
        self.agent = agent
        self._data = market

    @property
    def data(self):
        """
        Returns the MarketData for the Market

        Equivalent to calling markets() on an existing Market

        If a Ship or Probe is located at the Market, the live MarketData
        will be returned, including recent transactions, sell prices, and
        purchase prices

        Returns:
            MarketData
        """
        # REMINDER: This is the same data returned by Markets.__call__
        response = self.agent.client.get(
            f'/systems/{self.location.system}/waypoints/{self.symbol}/market'
        )
        data = response.json()['data']
        data['location'] = self.location
        if not data.get('tradeGoods'):
            data['tradeGoods'] = []
        return MarketData(self.agent, data)

    @property
    def location(self):
        """Returns the Market's Location"""
        return Location(self.agent, {'headquarters': self.symbol})


class MarketData(utils.AbstractJSONItem):

    """
    Convenience class. Equivalent to Market but useful for comparing Market
    to Market.data as Market will not have the supplemental data that
    MarketData contains

    >>> ship = next(iter(agent.fleet))
    >>> market = next(iter(ship.markets))
    >>> market.trade_goods  # Does not exist. Returns None
    >>> market.data.trade_goods  # or, market().trade_goods
    []
    """

    def __init__(self, agent, data):
        self.agent = agent
        self._data = data

    @property
    def data(self):
        """
        Returns the MarketData for the Market

        Equivalent to calling markets() on an existing Market

        If a Ship or Probe is located at the Market, the live MarketData
        will be returned, including recent transactions, sell prices, and
        purchase prices

        Returns:
            MarketData
        """
        # REMINDER: This is the same data returned by Markets.__call__
        response = self.agent.client.get(
            f'/systems/{self.location.system}/waypoints/{self.symbol}/market'
        )
        data = response.json()['data']
        data['location'] = self.location
        if not data.get('tradeGoods'):
            data['tradeGoods'] = []
        return MarketData(self.agent, data)

    @property
    def location(self):
        """Returns the Market's Location"""
        return Location(self.agent, {'headquarters': self.symbol})


def best_market_pairs(ship, market_data, price_delta=0):
    """
    Returns a sorted list of the best market pairs


    Args:
        ship: A ship in the System to search
        market_data: An iterable of MarketDatas to search

    Kwargs:
        price_delta: Minimum delta between the Purchase and Sell prices
                     Default is 0

    Returns:
        MarketDataRecord: Sorted list of MarketDataRecords if matches are
        found; else, an empty list
    """
    market_pairs = {}
    for market in market_data:
        for good in market.trade_goods:
            market_pairs.setdefault(good.symbol, []).append((market, good))
    output = []
    cache = {}

    for trade_symbol, markets in market_pairs.items():
        e_market, e_good = min(markets, key=lambda x: x[1].purchase_price)
        i_market, i_good = max(markets, key=lambda x: x[1].sell_price)
        if i_good.sell_price - e_good.purchase_price < price_delta:
            continue
        i_wp = cache.get(
            (
                i_market.location.system,
                i_market.location.waypoint
            )
        )
        if not i_wp:
            i_wp = ship.waypoints.get(
                system_symbol=i_market.location.system,
                waypoint_symbol=i_market.location.waypoint,
            )
            cache[
                (
                    i_market.location.system,
                    i_market.location.waypoint
                )
            ] = i_wp
        e_wp = cache.get(
            (
                e_market.location.system,
                e_market.location.waypoint
            )
        )
        if not e_wp:
            e_wp = ship.waypoints.get(
                system_symbol=e_market.location.system,
                waypoint_symbol=e_market.location.waypoint,
            )
            cache[
                (
                    e_market.location.system,
                    e_market.location.waypoint
                )
            ] = e_wp
        output.append(
            MarketDataRecord(
                distance=utils.calculate_waypoint_distance(i_wp, e_wp),
                trade_symbol=trade_symbol,
                import_market=i_market,
                export_market=e_market,
                import_waypoint=i_wp,
                export_waypoint=e_wp,
            )
        )
    return sorted(output, key=market_delta_sort_key, reverse=True)


def market_delta_sort_key(market_record):
    """
    Returns the difference between the Sell Price and Purchase Price
    for the trade goods associated with the market_record.trade_symbol

    Args:
        market_record: A MarketDataRecord
    """
    for i_good in market_record.import_market.trade_goods:
        if i_good.symbol == market_record.trade_symbol:
            for e_good in market_record.export_market.trade_goods:
                if e_good.symbol == market_record.trade_symbol:
                    return i_good.sell_price - e_good.purchase_price


def import_sort_key(market_data, trade_symbol):
    """
    Args:
        market_data: The Market's MarketData
        trade_symbol: Symbol of the good to check

    Returns:
        purchase_price: Returns sell price of the trade good if it imports
                        the trade_symbol
    """
    if market_data.trade_goods:
        for good in market_data.trade_goods:
            if good.type == 'IMPORT' and good.symbol == trade_symbol:
                return good.sell_price


def export_sort_key(market_data, trade_symbol):
    """
    Args:
        market_data: The Market's MarketData
        trade_symbol: Symbol of the good to check

    Returns:
        purchase_price: Returns purchase price of the trade good if it exports
                        the trade_symbol
    """
    if market_data.trade_goods:
        for good in market_data.trade_goods:
            if good.type == 'EXPORT' and good.symbol == trade_symbol:
                return good.purchase_price


def exchange_sort_key(market_data, trade_symbol):
    """
    Args:
        market_data: The Market's MarketData
        trade_symbol: Symbol of the good to check

    Returns:
        purchase_price: Returns purchase price of the trade good
                        if it exchanges the trade_symbol
    """
    if market_data.trade_goods:
        for good in market_data.trade_goods:
            if good.type == 'EXCHANGE' and good.symbol == trade_symbol:
                return good.purchase_price


def export_or_exchange_sort_key(market_data, trade_symbol):
    """
    Args:
        market_data: The Market's MarketData
        trade_symbol: Symbol of the good to check

    Returns:
        purchase_price: Returns purchase price of the trade good if it exports
                        or exchanges the trade_symbol
    """
    if market_data.trade_goods:
        for good in market_data.trade_goods:
            if good.symbol == trade_symbol:
                if good.type == 'EXPORT' or good.type == 'EXCHANGE':
                    return good.purchase_price


def has_import(market_data, trade_symbol):
    """
    Args:
        market_data: The Market's MarketData
        trade_symbol: Symbol of the good to check

    Returns:
        bool: True if market imports trade symbol; else, False
    """
    return trade_symbol in {i.symbol for i in market_data.imports}


def has_export(market_data, trade_symbol):
    """
    Args:
        market_data: The Market's MarketData
        trade_symbol: Symbol of the good to check

    Returns:
        bool: True if market exports trade symbol; else, False
    """
    return trade_symbol in {i.symbol for i in market_data.exports}


def has_exchange(market_data, trade_symbol):
    """
    Args:
        market_data: The Market's MarketData
        trade_symbol: Symbol of the good to check

    Returns:
        bool: True if market exchanges trade symbol; else, False
    """
    return trade_symbol in {i.symbol for i in market_data.exchange}



import httpx
import json
import os
import pytest


import snisp

from . import DATA_DIR


class TestCache:

    agent = snisp.agent.Agent(symbol='testing', faction='testing')
    client = agent.client
    client.testing = False

    def test_in_test(self, respx_mock):
        assert self.client.testing is False

    @pytest.mark.respx(base_url='https://api.spacetraders.io/v2')
    def test_get_fuel_stations(self, respx_mock):
        ship_data = json.load(
            open(os.path.join(DATA_DIR, 'ship_info.json'), encoding='utf8')
        )
        ship_data['data']['nav']['systemSymbol'] = 'TEST-SYSTEM'
        ship_data['data']['nav']['waypointSymbol'] = 'TEST-SYSTEM-WAYPOINT'
        respx_mock.get('/my/ships/TEST_SHIP_SYMBOL').mock(
            return_value=httpx.Response(200, json=ship_data)
        )

        fuel_stations_data = json.load(
            open(os.path.join(DATA_DIR, 'fuel_stations.json'), encoding='utf8')
        )
        respx_mock.get(
            '/systems/TEST-SYSTEM/waypoints', params={'page': 1}
        ).mock(
            return_value=httpx.Response(200, json=fuel_stations_data)
        )
        respx_mock.get(
            '/systems/TEST-SYSTEM/waypoints', params={'page': 2}
        ).mock(
            return_value=httpx.Response(200, json={'data': []})
        )

        market_data = json.load(
            open(os.path.join(DATA_DIR, 'market_data.json'), encoding='utf8')
        )
        respx_mock.get(
            '/systems/TEST-SYSTEM/waypoints/TEST-SYSTEM-NEXTCLOSESTFUEL/market'
        ).mock(
            return_value=httpx.Response(200, json=market_data)
        )
        respx_mock.get(
            '/systems/TEST-SYSTEM/waypoints/TEST-SYSTEM-FARTHESTFUEL/market'
        ).mock(
            return_value=httpx.Response(200, json=market_data)
        )
        respx_mock.get(
            '/systems/TEST-SYSTEM/waypoints/TEST-SYSTEM-CLOSESTFUEL/market'
        ).mock(
            return_value=httpx.Response(200, json=market_data)
        )

        ship = self.agent.fleet('TEST_SHIP_SYMBOL')
        fuel_stations = list(ship.markets.fuel_stations())
        assert fuel_stations == list(ship.markets.fuel_stations())
